""" Implementation of a step
"""
# pylint: disable=unused-wildcard-import,wildcard-import

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import os
import time
import traceback
import sys

from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


class StepDefinition(AttributeCollection):
    def __init__(self, step_type=None, obj=None, parent=None):
        AttributeCollection.__init__(self)
        self.module_instances = None
        # this object can be shared with the workflow
        # to if its =None and workflow!=None we have to
        # pass everything above
        self.module_definitions = None
        self.parent = parent

        # sort out Parameters and class attributes
        if (obj is None) or isinstance(obj, ParameterCollection):
            self.setType("notgiven")
            self.setDescrShort("")
            self.setDescription("")
            self.setOrigin("")
            self.setVersion(0.0)
            self.parameters = ParameterCollection(obj)  # creating copy
            self.module_instances = InstancesPool(self)
            self.module_definitions = DefinitionsPool(self)
        elif isinstance(obj, StepDefinition):
            self.setType(obj.getType())
            self.setDescrShort(obj.getDescrShort())
            self.setDescription(obj.getDescription())
            self.setOrigin(obj.getOrigin())
            self.setVersion(obj.getVersion())
            # copy instances and definitions
            self.parameters = ParameterCollection(obj.parameters)
            self.module_instances = InstancesPool(self, obj.module_instances)
            if obj.module_definitions is not None:
                self.module_definitions = DefinitionsPool(obj.module_definitions)
        else:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(obj)))
        if step_type:
            self.setType(step_type)

    def __str__(self):
        ret = str(type(self)) + ":\n" + AttributeCollection.__str__(self) + self.parameters.__str__()
        if self.module_definitions is not None:
            ret = ret + str(self.module_definitions)
        else:
            ret = ret + "Module definitions shared in Workflow\n"
        ret = ret + str(self.module_instances)
        return ret

    def toXML(self):
        ret = "<StepDefinition>\n"
        ret = ret + AttributeCollection.toXML(self)
        ret = ret + self.parameters.toXML()
        if self.module_definitions is not None:
            ret = ret + self.module_definitions.toXML()
        ret = ret + self.module_instances.toXML()
        ret = ret + "</StepDefinition>\n"
        return ret

    def toXMLFile(self, outFile):
        if os.path.exists(outFile):
            os.remove(outFile)
        with open(outFile, "w") as xmlfile:
            xmlfile.write(self.toXML())

    def addModule(self, module):
        # KGG We need to add code to update existing modules
        if self.module_definitions is None:
            self.parent.module_definitions.append(module)
        else:
            self.module_definitions.append(module)
        return module

    def createModuleInstance(self, module_type, name):
        """Creates module instance of type 'type' with the name 'name'"""

        if self.module_definitions[module_type]:
            mi = ModuleInstance(name, self.module_definitions[module_type])
            self.module_instances.append(mi)
            return mi
        else:
            raise KeyError("Can not find ModuleDefinition " + module_type + " to create ModuleInstrance " + name)

    def removeModuleInstance(self, name):
        """Remove module instance specified by its name"""
        self.module_instances.delete(name)

    def compare(self, s):
        """Custom Step comparison operation"""
        return AttributeCollection.compare(self, s) and self.module_instances.compare(s)

    def updateParent(self, parent):
        """ """
        # FIXME: no updateParents for AttributeCollection, What should this be?
        AttributeCollection.updateParents(self, parent)  # pylint: disable=no-member
        self.module_instances.updateParents(self)
        if self.module_definitions is not None:
            self.module_definitions.updateParents(self)

    def createCode(self):
        """Create Step code"""

        str = "class " + self.getType() + ":\n"
        str = str + indent(1) + "def execute(self):\n"
        str = str + self.module_instances.createCode()
        str = str + indent(2) + "# output assignment\n"
        for v in self.parameters:
            if v.isOutput():
                str = str + v.createParameterCode(2, "self")
        str += "\n"
        return str


class StepInstance(AttributeCollection):
    def __init__(self, name, obj=None, parent=None):
        AttributeCollection.__init__(self)
        self.parent = None

        if obj is None:
            self.parameters = ParameterCollection()
        elif isinstance(obj, StepInstance) or isinstance(obj, StepDefinition):
            if name is None:
                self.setName(obj.getName())
            else:
                self.setName(name)
            self.setType(obj.getType())
            self.setDescrShort(obj.getDescrShort())
            self.parameters = ParameterCollection(obj.parameters)
        elif (obj is None) or isinstance(obj, ParameterCollection):
            # set attributes
            self.setName(name)
            self.setType("")
            self.setDescrShort("")
            self.parameters = ParameterCollection(obj)
        elif obj is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(obj)))

        self.step_commons = {}
        self.stepStatus = S_OK()

    def resolveGlobalVars(self, step_definitions, wf_parameters):
        """Resolve parameter values defined in the @{<variable>} form"""
        self.parameters.resolveGlobalVars(wf_parameters)
        module_instance_number = 0
        for inst in step_definitions[self.getType()].module_instances:
            module_instance_number = module_instance_number + 1
            if not inst.parameters.find("MODULE_NUMBER"):
                inst.parameters.append(
                    Parameter(
                        "MODULE_NUMBER",
                        "%s" % module_instance_number,
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "ModuleInstance number within the Step",
                    )
                )
            if not inst.parameters.find("MODULE_INSTANCE_NAME"):
                inst.parameters.append(
                    Parameter(
                        "MODULE_INSTANCE_NAME",
                        inst.getName(),
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Name of the ModuleInstance within the Step",
                    )
                )
            if not inst.parameters.find("MODULE_DEFINITION_NAME"):
                inst.parameters.append(
                    Parameter(
                        "MODULE_DEFINITION_NAME",
                        inst.getType(),
                        "string",
                        "",
                        "",
                        True,
                        False,
                        "Type of the ModuleInstance within the Step",
                    )
                )
            if not inst.parameters.find("JOB_ID"):
                inst.parameters.append(
                    Parameter(
                        "JOB_ID", "", "string", "self", "JOB_ID", True, False, "Job ID within a Production as a string"
                    )
                )
            if not inst.parameters.find("PRODUCTION_ID"):
                inst.parameters.append(
                    Parameter(
                        "PRODUCTION_ID", "", "string", "self", "PRODUCTION_ID", True, False, "Production ID as a string"
                    )
                )
            if not inst.parameters.find("STEP_NUMBER"):
                inst.parameters.append(
                    Parameter(
                        "STEP_NUMBER",
                        "",
                        "string",
                        "self",
                        "STEP_NUMBER",
                        True,
                        False,
                        "Step instance number within the Workflow",
                    )
                )
            if not inst.parameters.find("STEP_ID"):
                inst.parameters.append(
                    Parameter(
                        "STEP_ID", "", "string", "self", "STEP_NUMBER", True, False, "Step ID within the Workflow"
                    )
                )
            inst.resolveGlobalVars(wf_parameters, self.parameters)

    def createCode(self, ind=2):
        """Create the Step code"""
        str = indent(ind) + self.getName() + " = " + self.getType() + "()\n"
        str = str + self.parameters.createParametersCode(ind, self.getName())
        str = str + indent(ind) + self.getName() + ".execute()\n\n"
        return str

    def __str__(self):
        """Step string representation"""
        return str(type(self)) + ":\n" + AttributeCollection.__str__(self) + self.parameters.__str__()

    def toXML(self):
        """Generate the Step XML representation"""
        ret = "<StepInstance>\n"
        ret = ret + AttributeCollection.toXML(self)
        ret = ret + self.parameters.toXML()
        ret = ret + "</StepInstance>\n"
        return ret

    def setWorkflowCommons(self, wf):
        """Add reference to the collection of the common tools"""

        self.workflow_commons = wf

    def execute(self, step_exec_attr, definitions):
        """Step execution method. step_exec_attr is array to hold parameters belong to this Step,
        filled above in the workflow
        """
        print("Executing StepInstance", self.getName(), "of type", self.getType(), list(definitions))
        # Report the Application state if the corresponding tool is supplied
        if "JobReport" in self.workflow_commons:
            if self.parent.workflowStatus["OK"]:
                self.workflow_commons["JobReport"].setApplicationStatus("Executing " + self.getName())

        # Prepare Step statistics evaluation
        self.step_commons["StartTime"] = time.time()
        self.step_commons["StartStats"] = os.times()

        step_def = definitions[self.getType()]
        step_exec_modules = {}
        error_message = ""
        error_code = 0
        for mod_inst in step_def.module_instances:
            mod_inst_name = mod_inst.getName()
            mod_inst_type = mod_inst.getType()

            # print "StepInstance creating module instance ", mod_inst_name, " of type", mod_inst.getType()
            step_exec_modules[mod_inst_name] = step_def.parent.module_definitions[
                mod_inst_type
            ].main_class_obj()  # creating instance

            # Resolve all the linked parameter values
            for parameter in mod_inst.parameters:
                if parameter.preExecute():
                    # print '>>>> Input', parameter
                    if parameter.isLinked():
                        # print ">>>> ModuleInstance", mod_inst_name + '.' + parameter.getName(),
                        # '=', parameter.getLinkedModule() + '.' + parameter.getLinkedParameter()
                        if parameter.getLinkedModule() == "self":
                            # tale value form the step_dict
                            setattr(
                                step_exec_modules[mod_inst_name],
                                parameter.getName(),
                                step_exec_attr[parameter.getLinkedParameter()],
                            )
                        else:

                            setattr(
                                step_exec_modules[mod_inst_name],
                                parameter.getName(),
                                getattr(step_exec_modules[parameter.getLinkedModule()], parameter.getLinkedParameter()),
                            )
                    else:
                        # print ">>>> ModuleInstance", mod_inst_name + '.' + parameter.getName(), '=', parameter.getValue()
                        setattr(step_exec_modules[mod_inst_name], parameter.getName(), parameter.getValue())
                    # print 'Step Input Parameter:', parameter.getName(), getattr(
                    # step_exec_modules[mod_inst_name], parameter.getName() )

            # Set reference to the workflow and step common tools
            setattr(step_exec_modules[mod_inst_name], "workflow_commons", self.parent.workflow_commons)
            setattr(step_exec_modules[mod_inst_name], "step_commons", self.step_commons)
            setattr(step_exec_modules[mod_inst_name], "stepStatus", self.stepStatus)
            setattr(step_exec_modules[mod_inst_name], "workflowStatus", self.parent.workflowStatus)

            try:
                result = step_exec_modules[mod_inst_name].execute()
                if not result["OK"]:
                    if self.stepStatus["OK"]:
                        error_message = result["Message"]
                        error_code = result["Errno"]
                        if "JobReport" in self.workflow_commons:
                            if self.parent.workflowStatus["OK"]:
                                self.workflow_commons["JobReport"].setApplicationStatus(error_message)
                    self.stepStatus = S_ERROR(result["Message"])
                else:
                    for parameter in mod_inst.parameters:
                        if parameter.isOutput():
                            # print '<<<< Output', parameter
                            if parameter.isLinked():
                                # print "ModuleInstance self ." + parameter.getName(), '=',
                                # parameter.getLinkedModule() + '.' + parameter.getLinkedParameter()
                                if parameter.getLinkedModule() == "self":
                                    # this is not supposed to happen
                                    print("Warning! Module OUTPUT attribute", parameter.getName(), end=" ")
                                    print(
                                        "refer to the attribute of the same module",
                                        parameter.getLinkedParameter(),
                                        "=",
                                        getattr(step_exec_modules[mod_inst_name], parameter.getName()),
                                    )
                                    step_exec_attr[parameter.getName()] = getattr(
                                        step_exec_modules[mod_inst_name],
                                        parameter.getLinkedParameter(),
                                        parameter.getValue(),
                                    )
                                    # print "                 OUT", parameter.getLinkedParameter(), '=',
                                    # getattr( step_exec_modules[mod_inst_name], parameter.getName(),
                                    # parameter.getValue() )
                                else:
                                    # print 'Output step_exec_attr', st_parameter.getName(),
                                    # step_exec_modules[st_parameter.getLinkedModule()],
                                    # parameter.getLinkedParameter()
                                    step_exec_attr[parameter.getName()] = getattr(
                                        step_exec_modules[parameter.getLinkedModule()], parameter.getLinkedParameter()
                                    )
                            else:
                                # This also does not make sense - we can give a warning
                                print("Warning! Module OUTPUT attribute ", parameter.getName(), end=" ")
                                print("assigned constant", parameter.getValue())
                                # print "StepInstance self." + parameter.getName(), '=', parameter.getValue()
                                step_exec_attr[parameter.getName()] = parameter.getValue()

                            # print 'Module Output Parameter:', parameter.getName(), step_exec_attr[parameter.getName()]

                    # Get output values to the step_commons dictionary
                    for key in result:
                        if key != "OK":
                            if key != "Value":
                                self.step_commons[key] = result[key]
                            elif isinstance(result["Value"], dict):
                                for vkey in result["Value"].keys():
                                    self.step_commons[vkey] = result["Value"][vkey]

            except Exception as x:
                print("Exception while module execution")
                print("Module", mod_inst_name, mod_inst.getType())
                print(str(x))
                exc = sys.exc_info()
                exc_type = exc[0]
                value = exc[1]
                print(
                    "== EXCEPTION ==\n%s: %s\n\n%s==============="
                    % (exc_type, value, "\n".join(traceback.format_tb(exc[2])))
                )

                print("Step status: ", self.stepStatus)
                print("Workflow status: ", self.parent.workflowStatus)
                if self.stepStatus["OK"]:
                    # This is the error that caused the workflow disruption
                    # report it to the WMS
                    error_message = "Exception while %s module execution: %s" % (mod_inst_name, str(x))
                    error_code = 0
                    if "JobReport" in self.workflow_commons:
                        if self.parent.workflowStatus["OK"]:
                            self.workflow_commons["JobReport"].setApplicationStatus(
                                "Exception in %s module" % mod_inst_name
                            )

                self.stepStatus = S_ERROR(error_code, error_message)

        # now we need to copy output values to the STEP!!! parameters
        for st_parameter in self.parameters:
            if st_parameter.isOutput():
                # print '<< Output', st_parameter
                if st_parameter.isLinked():
                    # print "StepInstance self." + st_parameter.getName(), '=',
                    # st_parameter.getLinkedModule() + '.' + st_parameter.getLinkedParameter()
                    if st_parameter.getLinkedModule() == "self":
                        # this is not supposed to happen
                        print("Warning! Step OUTPUT attribute", st_parameter.getName(), end=" ")
                        print(
                            "refer to the attribute of the same step",
                            st_parameter.getLinkedParameter(),
                            step_exec_attr[st_parameter.getLinkedParameter()],
                        )
                        step_exec_attr[st_parameter.getName()] = step_exec_attr[st_parameter.getLinkedParameter()]

                    else:
                        # print 'Output step_exec_attr', st_parameter.getName(),
                        # step_exec_modules[st_parameter.getLinkedModule()],
                        # st_parameter.getLinkedParameter()
                        step_exec_attr[st_parameter.getName()] = getattr(
                            step_exec_modules[st_parameter.getLinkedModule()], st_parameter.getLinkedParameter()
                        )
                    setattr(self, st_parameter.getName(), step_exec_attr[st_parameter.getName()])
                else:
                    # This also does not make sense - we can give a warning
                    print("Warning! Step OUTPUT attribute ", st_parameter.getName(), end=" ")
                    print("assigned constant", st_parameter.getValue())
                    # print "StepInstance self." + st_parameter.getName(), '=', st_parameter.getValue()
                    step_exec_attr[st_parameter.getName()] = st_parameter.getValue()

                print("Step Output", st_parameter.getName(), "=", step_exec_attr[st_parameter.getName()])

        # Return the result of the first failed module or S_OK
        if not self.stepStatus["OK"]:
            return S_ERROR(error_code, error_message)
        else:
            return S_OK(result["Value"])



""" Accounting reporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hashlib
import re

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities.ObjectLoader import loadObjects
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
from DIRAC.AccountingSystem.private.Policies import gPoliciesList
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter as myBaseReporter


class PlottersList(object):
    def __init__(self):
        objectsLoaded = loadObjects(
            "AccountingSystem/private/Plotters", re.compile(r".*[a-z1-9]Plotter\.py$"), myBaseReporter
        )
        self.__plotters = {}
        for objName in objectsLoaded:
            self.__plotters[objName[:-7]] = objectsLoaded[objName]

    def getPlotterClass(self, typeName):
        try:
            return self.__plotters[typeName]
        except KeyError:
            return None


class MainReporter(object):
    def __init__(self, db, setup):
        self._db = db
        self.setup = setup
        self.csSection = getServiceSection("Accounting/ReportGenerator", setup=setup)
        self.plotterList = PlottersList()

    def __calculateReportHash(self, reportRequest):
        requestToHash = dict(reportRequest)
        granularity = gConfig.getValue("%s/CacheTimeGranularity" % self.csSection, 300)
        for key in ("startTime", "endTime"):
            epoch = requestToHash[key]
            requestToHash[key] = epoch - epoch % granularity
        md5Hash = hashlib.md5()
        md5Hash.update(repr(requestToHash).encode())
        md5Hash.update(self.setup.encode())
        return md5Hash.hexdigest()

    def generate(self, reportRequest, credDict):
        typeName = reportRequest["typeName"]
        plotterClass = self.plotterList.getPlotterClass(typeName)
        if not plotterClass:
            return S_ERROR("There's no reporter registered for type %s" % typeName)
        if typeName in gPoliciesList:
            retVal = gPoliciesList[typeName].checkRequest(
                reportRequest["reportName"], credDict, reportRequest["condDict"], reportRequest["grouping"]
            )
            if not retVal["OK"]:
                return retVal
        reportRequest["hash"] = self.__calculateReportHash(reportRequest)
        plotter = plotterClass(self._db, self.setup, reportRequest["extraArgs"])
        return plotter.generate(reportRequest)

    def list(self, typeName):
        plotterClass = self.plotterList.getPlotterClass(typeName)
        if not plotterClass:
            return S_ERROR("There's no plotter registered for type %s" % typeName)
        plotter = plotterClass(self._db, self.setup)
        return S_OK(plotter.plotsList())



"""
    This is a comment
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# $Source: /tmp/libdirac/tmp.stZoy15380/dirac/DIRAC3/DIRAC/Core/Workflow/test/JobSamples.py,v $

from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC.Core.Workflow.Step import *
from DIRAC.Core.Workflow.Workflow import *
from DIRAC.Core.Workflow.WorkflowReader import *

# define Module 2
module2 = ModuleDefinition("GaudiApplication")  # during constraction class creates duplicating copies of the params
module2.setDescription("Gaudi Application module")
module2.setBody("from WorkflowLib.Module.GaudiApplication import GaudiApplication\n")

# we add empty parameters but linked up as default
module2.addParameter(Parameter("DataType", "", "string", "self", "DataType", True, False, "data type"))
module2.addParameter(Parameter("CONFIG_NAME", "", "string", "self", "CONFIG_NAME", True, False, "Configuration Name"))
module2.addParameter(
    Parameter("CONFIG_VERSION", "", "string", "self", "CONFIG_VERSION", True, False, "Configuration Version")
)
module2.addParameter(Parameter("EVENTTYPE", "", "string", "self", "EVENTTYPE", True, False, "Event Type"))
module2.addParameter(Parameter("appName", "", "string", "self", "appName", True, False, "Application Name"))
module2.addParameter(Parameter("appVersion", "", "string", "self", "appVersion", True, False, "Application Version"))
module2.addParameter(Parameter("appType", "", "string", "self", "appType", True, False, "Application Version"))
module2.addParameter(Parameter("appLog", "", "string", "self", "appLog", True, False, "list of logfile"))
module2.addParameter(Parameter("inputData", "", "jdl", "self", "inputData", True, False, "List of InputData"))
module2.addParameter(Parameter("inputDataType", "", "string", "self", "inputDataType", True, False, "Input Data Type"))
# module2.addParameter(Parameter(
# "nb_events_input","","string","self","nb_events_input",True,False,"number of events as input"))
module2.addParameter(
    Parameter("NUMBER_OF_EVENTS", "", "string", "self", "NUMBER_OF_EVENTS", True, False, "number of events requested")
)
module2.addParameter(Parameter("optionsFile", "", "string", "self", "optionsFile", True, False, "Options File"))
module2.addParameter(
    Parameter("optionsLine", "", "string", "self", "optionsLine", True, False, "Number of Event", "option")
)
module2.addParameter(Parameter("systemConfig", "", "string", "self", "systemConfig", True, False, "Job Platform"))
module2.addParameter(Parameter("poolXMLCatName", "", "string", "self", "poolXMLCatName", True, False, "POOL XML slice"))
module2.addParameter(Parameter("outputData", "", "string", "self", "outputData", True, False, "list of output data"))

# define module 3
module3 = ModuleDefinition("LogChecker")  # during constraction class creates duplicating copies of the params
module3.setDescription("Check LogFile module")
module3.setBody("from WorkflowLib.Module.LogChecker import *\n")

# we add parameters and link them to the level of step
module3.addParameter(Parameter("DataType", "", "string", "self", "DataType", True, False, "data type"))
module3.addParameter(Parameter("CONFIG_NAME", "", "string", "self", "CONFIG_NAME", True, False, "Configuration Name"))
module3.addParameter(
    Parameter("CONFIG_VERSION", "", "string", "self", "CONFIG_VERSION", True, False, "Configuration Version")
)
module3.addParameter(
    Parameter("NUMBER_OF_EVENTS", "", "string", "self", "NUMBER_OF_EVENTS", True, False, "number of events requested")
)
# module3.addParameter(
# Parameter("nb_events_input","","string","self","nb_events_input",True,False,"number of events as input"))
module3.addParameter(Parameter("appName", "", "string", "self", "appName", True, False, "Application Name"))
module3.addParameter(Parameter("appVersion", "", "string", "self", "appVersion", True, False, "Application Version"))
module3.addParameter(Parameter("appType", "", "string", "self", "appType", True, False, "Application Version"))
module3.addParameter(Parameter("appLog", "", "string", "self", "appLog", True, False, "list of logfile"))
module3.addParameter(Parameter("poolXMLCatName", "", "string", "self", "poolXMLCatName", True, False, "POOL XML slice"))
module3.addParameter(Parameter("inputData", "", "string", "self", "inputData", True, False, "InputData"))
module3.addParameter(
    Parameter("OUTPUT_MAX", "", "string", "self", "OUTPUT_MAX", True, False, "nb max of output to keep")
)
# this parameter is static so we define it here
module3.addParameter(Parameter("EMAIL", "@{EMAILNAME}", "string", "", "", True, False, "EMAIL adress"))
module3.addParameter(Parameter("outputDataSE", "", "string", "self", "outputDataSE", True, False, "SE of output data"))


# define module 4
module4 = ModuleDefinition("BookkeepingReport")
module4.setDescription("Bookkeeping Report module")
module4.setBody("from WorkflowLib.Module.BookkeepingReport import * \n")
module4.addParameter(Parameter("STEP_ID", "", "string", "self", "STEP_ID", True, False, "EMAIL adress"))
# module4.addParameter(
# Parameter("nb_events_input","","string","self","nb_events_input",True,False,"number of events as input"))
module4.addParameter(
    Parameter("NUMBER_OF_EVENTS", "", "string", "self", "NUMBER_OF_EVENTS", True, False, "number of events requested")
)
module4.addParameter(Parameter("DataType", "", "string", "self", "DataType", True, False, "data type"))
module4.addParameter(Parameter("CONFIG_NAME", "", "string", "self", "CONFIG_NAME", True, False, "Configuration Name"))
module4.addParameter(
    Parameter("CONFIG_VERSION", "", "string", "self", "CONFIG_VERSION", True, False, "Configuration Version")
)
module4.addParameter(Parameter("appName", "", "string", "self", "appName", True, False, "Application Name"))
module4.addParameter(Parameter("appVersion", "", "string", "self", "appVersion", True, False, "Application Version"))
module4.addParameter(Parameter("inputData", "", "string", "self", "inputData", True, False, "InputData"))
module4.addParameter(Parameter("inputDataType", "", "string", "self", "inputDataType", True, False, "Input Data Type"))
module4.addParameter(Parameter("EVENTTYPE", "", "string", "self", "EVENTTYPE", True, False, "Event Type"))
module4.addParameter(Parameter("outputData", "", "string", "self", "outputData", True, False, "list of output data"))
module4.addParameter(Parameter("appType", "", "string", "self", "appType", True, False, "Application Version"))
module4.addParameter(Parameter("poolXMLCatName", "", "string", "self", "poolXMLCatName", True, False, "POOL XML slice"))
module4.addParameter(Parameter("appLog", "", "string", "self", "appLog", True, False, "list of logfile"))


# define module 5
module5 = ModuleDefinition("StepFinalization")
module5.setDescription("Step Finalization module")
module5.setBody("from WorkflowLib.Module.StepFinalization import * \n")
module5.addParameter(Parameter("outputDataSE", "", "string", "self", "outputDataSE", True, False, "SE of output data"))
module5.addParameter(Parameter("outputData", "", "string", "self", "outputData", True, False, "list of output data"))
module5.addParameter(Parameter("poolXMLCatName", "", "string", "self", "poolXMLCatName", True, False, "POOL XML slice"))
module5.addParameter(Parameter("inputData", "", "string", "self", "inputData", True, False, "InputData"))
module5.addParameter(Parameter("appType", "", "string", "self", "appType", True, False, "Application Version"))

# define module 6
module6 = ModuleDefinition("JobFinalization")
module6.setDescription("Job Finalization module")
module6.setBody("from WorkflowLib.Module.JobFinalization import * \n")
module6.addParameter(Parameter("outputDataSE", "", "string", "self", "outputDataSE", True, False, "SE of output data"))
module6.addParameter(Parameter("outputData", "", "string", "self", "outputData", True, False, "list of output data"))
module6.addParameter(Parameter("poolXMLCatName", "", "string", "self", "poolXMLCatName", True, False, "POOL XML slice"))
module6.addParameter(Parameter("inputData", "", "string", "self", "inputData", True, False, "InputData"))
module6.addParameter(Parameter("appType", "", "string", "self", "appType", True, False, "Application Version"))


# ##############   STEPS ##################################
# step 1 we are creating step definition
step1 = StepDefinition("Gaudi_App_Step")
step1.addModule(module2)  # Creating instance of the module 'Gaudi_App_Step'
moduleInstance2 = step1.createModuleInstance("GaudiApplication", "module2")

step1.addModule(module3)  # Creating instance of the module 'LogChecker'
moduleInstance3 = step1.createModuleInstance("LogChecker", "module3")

step1.addModule(module4)  # Creating instance of the module 'LogChecker'
moduleInstance4 = step1.createModuleInstance("BookkeepingReport", "module4")
# in principle we can link parameters of moduleInstance2 with moduleInstance1 but
# in this case we going to use link with the step
step1.addModule(module5)
moduleInstance5 = step1.createModuleInstance("StepFinalization", "module5")
# moduleInstance5.addParameter(
# Parameter("outputData_1",'',"string","self","outputData_1",True,False,"list of output data"))
# moduleInstance5.addParameter(
# Parameter("appType_1","","string","self","appType_1",True,False,"Application Version"))
# moduleInstance5.addParameter(
# Parameter("outputDataSE_1","","string","self","outputDataSE_1",True,False,"SE of output data"))

# and we can add additional parameter which will be used as a global
step1.addParameter(
    Parameter("STEP_ID", "@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}", "string", "", "", True, False, "Temporary fix")
)
step1.addParameter(Parameter("EVENTTYPE", "30000000", "string", "", "", True, False, "Event Type"))
step1.addParameter(Parameter("outputData", "@{STEP_ID}.root", "string", "", "", True, False, "etc name"))
step1.addParameter(Parameter("outputDataSE", "Tier1-RDST", "string", "", "", True, False, "etc name"))
step1.addParameter(Parameter("etcf", "SETC_@{STEP_ID}.root", "string", "", "", True, False, "etc name"))
step1.setValue("appLog", "@{appName}_@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}.log")

# outputData = "@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}.@{appType}"
opt_dav = "EvtTupleSvc.Output = {}"
opt_dav = opt_dav + ";ApplicationMgr.OutStream -= {'DstWriter'}"
opt_dav = opt_dav + ";ApplicationMgr.OutStream = {'Sequencer/SeqWriteTag'}"
opt_dav = opt_dav + ';ApplicationMgr.TopAlg -= { "GaudiSequencer/SeqPreselHWZ2bbl" }'
opt_dav = opt_dav + ";MessageSvc.Format = '%u % F%18W%S%7W%R%T %0W%M';MessageSvc.timeFormat = '%Y-%m-%d %H:%M:%S UTC'"
opt_dav = (
    opt_dav
    + ";WR.Output = \"Collection='EVTTAGS/TagCreator/1' ADDRESS='/Event' "
    + "DATAFILE='@{outputData}' TYP='POOL_ROOTTREE' OPT='RECREATE'\""
)
opt_dav = opt_dav + ';EventPersistencySvc.CnvServices += { "LHCb::RawDataCnvSvc" }'
opt_dav = opt_dav + ';ApplicationMgr.TopAlg += {"StoreExplorerAlg"}'
opt_dav = opt_dav + ";StoreExplorerAlg.Load = 1"
opt_dav = opt_dav + ";StoreExplorerAlg.PrintFreq = 0.99"
opt_dav = opt_dav + ";StoreExplorerAlg.AccessForeign = true"

# etcf = "SETC_@{STEP_ID}.root"
opt_brunel = '#include "$BRUNELOPTS/SuppressWarnings.opts"'
opt_brunel = (
    opt_brunel + ";MessageSvc.Format = '%u % F%18W%S%7W%R%T %0W%M';" + "MessageSvc.timeFormat = '%Y-%m-%d %H:%M:%S UTC'"
)
opt_brunel = opt_brunel + ";EventLoopMgr.OutputLevel = 3"
opt_brunel = opt_brunel + ";DstWriter.Output = \"DATAFILE='PFN:@{outputData}' TYP='POOL_ROOTTREE' OPT='RECREATE'\""
opt_brunel = (
    opt_brunel + ";EvtTupleSvc.Output = {\"EVTTAGS2 DATAFILE='PFN:@{etcf}' TYP='POOL_ROOTTREE' OPT='RECREATE'\"}"
)
# opt_brunel = opt_brunel+
# ";EventSelector.Input = {\"COLLECTION=\'TagCreator/1\' DATAFILE=\'@{InputData}\'
# TYPE=\'POOL_ROOTTREE\' SEL=\'(GlobalOr>=1)\' OPT=\'READ\'\"}"
opt_brunel = opt_brunel + ';EventPersistencySvc.CnvServices += { "LHCb::RawDataCnvSvc" }'
opt_brunel = opt_brunel + ';ApplicationMgr.TopAlg += {"StoreExplorerAlg"}'
opt_brunel = opt_brunel + ";StoreExplorerAlg.Load = 1"
opt_brunel = opt_brunel + ";StoreExplorerAlg.PrintFreq = 0.99"
opt_brunel = opt_brunel + ";IODataManager.AgeLimit = 2"

step3 = StepDefinition("Job_Finalization")
step3.addModule(module6)
moduleInstance6 = step3.createModuleInstance("JobFinalization", "module6")


##############  WORKFLOW #################################
workflow1 = Workflow(name="CCRC-joel-test")
workflow1.setDescription("Workflow of GaudiApplication")

workflow1.addStep(step1)
step1_prefix = "step1_"
stepInstance1 = workflow1.createStepInstance("Gaudi_App_Step", "Step1")
# lets link all parameters them up with the level of workflow
stepInstance1.setLink("systemConfig", "self", "SystemConfig")  # capital letter corrected
# except "STEP_ID", "appLog"
stepInstance1.setValue("appName", "DaVinci")
stepInstance1.setValue("appType", "root")
stepInstance1.setValue("outputData", "@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}.@{appType}")
stepInstance1.setValue("optionsFile", "DVOfficialStrippingFile.opts")
stepInstance1.setValue("optionsLine", opt_dav)
stepInstance1.setValue("outputDataSE", "Tier1_M-DST")
stepInstance1.setLink("inputData", "self", "InputData")  # KGG linked with InputData of the Workflow
# setpInstance1_module5 = stepInstance1.module_instances.find(module5)

step2_prefix = "step2_"
stepInstance2 = workflow1.createStepInstance("Gaudi_App_Step", "Step2")
# lets link all parameters them up with the level of workflow
stepInstance2.setLink("systemConfig", "self", "SystemConfig")  # capital letter corrected
# except "STEP_ID", "appLog"
stepInstance2.setValue("appName", "Brunel")
stepInstance2.setValue("appType", "dst")
stepInstance2.setValue("outputData", "@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}.@{appType}")
stepInstance2.setValue("etcf", "SETC_@{PRODUCTION_ID}_@{JOB_ID}_@{STEP_NUMBER}.root")
stepInstance2.setValue("optionsFile", "RealData-ETC.opts")
stepInstance2.setValue("optionsLine", opt_brunel)
stepInstance2.setValue("outputDataSE", "Tier1_M-DST")
stepInstance2.setLink("inputData", stepInstance1.getName(), "outputData")

workflow1.addStep(step3)
step3_prefix = "step3_"
stepInstance3 = workflow1.createStepInstance("Job_Finalization", "Step3")

# Now lets define parameters on the top
# indata = "LFN:/lhcb/production/DC06/phys-v2-lumi2/00001820/SIM/0000/00001820_00000001_1.sim;
# LFN:/lhcb/production/DC06/phys-v2-lumi2/00001820/SIM/0000/00001820_00000001_2.sim;
# LFN:/lhcb/production/DC06/phys-v2-lumi2/00001820/SIM/0000/00001820_00000001_3.sim"
# indata = "LFN:/lhcb/data/CCRC08/RDST/00000106/0000/00000106_00007918_1.rdst"
indata = (
    "LFN:/lhcb/data/CCRC08/RDST/00000106/0000/00000106_00007918_1.rdst;"
    + "/lhcb/data/CCRC08/RAW/LHCb/CCRC/420217/420217_0000116193.raw"
)
# etcf = "joel.root"
# indata = "LFN:/lhcb/data/CCRC08/RAW/LHCb/CCRC/402154/402154_0000047096.raw;
# LFN:/lhcb/data/CCRC08/RAW/LHCb/CCRC/402154/402154_0000047097.raw"

workflow1.setValue(step1_prefix + "appVersion", "v19r11")
workflow1.setValue(step2_prefix + "appVersion", "v32r4")
workflow1.setValue(step1_prefix + "NUMBER_OF_EVENTS", "5")
workflow1.setValue(step2_prefix + "NUMBER_OF_EVENTS", "-1")
workflow1.setValue(step1_prefix + "poolXMLCatName", "pool_xml_catalog.xml")
workflow1.setValue(step2_prefix + "poolXMLCatName", "pool_xml_catalog.xml")
# workflow1.setValue(step1_prefix+"inputData",indata)
workflow1.removeParameter(step1_prefix + "inputData")  # KGG wrong parameter
workflow1.setValue(step1_prefix + "inputDataType", "RDST")
workflow1.setValue(step2_prefix + "inputDataType", "ETC")
workflow1.setValue(step1_prefix + "OUTPUT_MAX", "20")
# remove unwanted
workflow1.removeParameter(step1_prefix + "outputData")
workflow1.removeParameter(step1_prefix + "systemConfig")
workflow1.removeParameter(step2_prefix + "outputData")
workflow1.removeParameter(step2_prefix + "systemConfig")
# add syspem config which common for all modules
# workflow1.addParameter(Parameter("systemConfig","slc4_ia32_gcc34","string","","",True, False, "Application Name"))
workflow1.addParameter(
    Parameter("SystemConfig", "x86_64-slc5-gcc43-opt", "JDLReqt", "", "", True, False, "Application Name")
)

workflow1.addParameter(Parameter("InputData", indata, "JDL", "", "", True, False, "Application Name"))
workflow1.addParameter(Parameter("JobType", "test", "JDL", "", "", True, False, "Job TYpe"))
workflow1.addParameter(Parameter("AncestorDepth", "2", "JDL", "", "", True, False, "Ancestor Depth"))
workflow1.addParameter(Parameter("Owner", "joel", "JDL", "", "", True, False, "user Name"))
workflow1.addParameter(Parameter("StdError", "std.err", "JDL", "", "", True, False, "user Name"))
workflow1.addParameter(Parameter("StdOutput", "std.out", "JDL", "", "", True, False, "user Name"))
workflow1.addParameter(Parameter("SoftwarePackages", "Brunel.v32r3p1", "JDL", "", "", True, False, "software"))

workflow1.addParameter(Parameter("CPUTime", 300000, "JDLReqt", "", "", True, False, "Application Name"))
# workflow1.addParameter(Parameter("Site","LCG.CERN.ch","JDLReqt","","",True, False, "Site"))
workflow1.addParameter(Parameter("Platform", "gLite", "JDLReqt", "", "", True, False, "platform"))

workflow1.addParameter(Parameter("PRODUCTION_ID", "00003033", "string", "", "", True, False, "Temporary fix"))
workflow1.addParameter(Parameter("JOB_ID", "00000011", "string", "", "", True, False, "Temporary fix"))
workflow1.addParameter(
    Parameter(
        "EMAILNAME",
        "joel.closier@cern.ch",
        "string",
        "",
        "",
        True,
        False,
        "Email to send a report from the LogCheck module",
    )
)
workflow1.addParameter(Parameter("DataType", "DATA", "string", "", "", True, False, "type of Datatype"))
workflow1.addParameter(Parameter("CONFIG_NAME", "LHCb", "string", "", "", True, False, "Configuration Name"))
workflow1.addParameter(Parameter("CONFIG_VERSION", "CCRC08", "string", "", "", True, False, "Configuration Version"))
# workflow1.addParameter(Parameter("NUMBER_OF_EVENTS","5","string","","",True, False, "number of events requested"))
workflow1.toXMLFile("wkf_CCRC_3.xml")
# w4 = fromXMLFile("/afs/cern.ch/user/g/gkuznets/test1.xml")
# print 'Creating code for the workflow'
# print workflow1.createCode()
# eval(compile(workflow1.createCode(),'<string>','exec'))
# workflow1.execute()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-show-task-queues
# Author :  Ricardo Graciani
########################################################################
"""
Show details of currently active Task Queues

Example:
  $ dirac-admin-show-task-queues
  Getting TQs..
  * TQ 401
          CPUTime: 360
             Jobs: 3
          OwnerDN: /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar
       OwnerGroup: dirac_user
         Priority: 1.0
            Setup: Dirac-Production
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.PrettyPrint import printTable

from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUsernameForDN
from DIRAC.WorkloadManagementSystem.Client.MatcherClient import MatcherClient
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

verbose = False


def setVerbose(optVal):
    global verbose
    verbose = True
    return S_OK()


taskQueueID = 0


def setTaskQueueID(optVal):
    global taskQueueID
    taskQueueID = int(optVal)
    return S_OK()


@Script()
def main():
    global verbose
    global taskQueueID
    Script.registerSwitch("v", "verbose", "give max details about task queues", setVerbose)
    Script.registerSwitch("t:", "taskQueue=", "show this task queue only", setTaskQueueID)
    Script.parseCommandLine(initializeMonitor=False)

    result = MatcherClient().getActiveTaskQueues()
    if not result["OK"]:
        gLogger.error(result["Message"])
        sys.exit(1)

    tqDict = result["Value"]

    if not verbose:
        fields = [
            "TaskQueue",
            "Jobs",
            "CPUTime",
            "Owner",
            "OwnerGroup",
            "Sites",
            "Platforms",
            "SubmitPools",
            "Setup",
            "Priority",
        ]
        records = []

        for tqId in sorted(tqDict):
            if taskQueueID and tqId != taskQueueID:
                continue
            record = [str(tqId)]
            tqData = tqDict[tqId]
            for key in fields[1:]:
                if key == "Owner":
                    value = tqData.get("OwnerDN", "-")
                    if value != "-":
                        result = getUsernameForDN(value)
                        if not result["OK"]:
                            value = "Unknown"
                        else:
                            value = result["Value"]
                else:
                    value = tqData.get(key, "-")
                if isinstance(value, list):
                    if len(value) > 1:
                        record.append(str(value[0]) + "...")
                    else:
                        record.append(str(value[0]))
                else:
                    record.append(str(value))
            records.append(record)

        printTable(fields, records)
    else:
        fields = ["Key", "Value"]
        for tqId in sorted(tqDict):
            if taskQueueID and tqId != taskQueueID:
                continue
            gLogger.notice("\n==> TQ %s" % tqId)
            records = []
            tqData = tqDict[tqId]
            for key in sorted(tqData):
                value = tqData[key]
                if isinstance(value, list):
                    records.append([key, {"Value": value, "Just": "L"}])
                else:
                    value = str(value)
                    records.append([key, {"Value": value, "Just": "L"}])

            printTable(fields, records, numbering=False)


if __name__ == "__main__":
    main()



########################################################################
# $HeadURL $
# File: ForwardDISETTests.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/04/18 09:23:05
########################################################################

""" :mod: ForwardDISETTests
    =======================

    .. module: ForwardDISETTests
    :synopsis: unittest for ForwardDISET handler
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    unittest for ForwardDISET handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file ForwardDISETTests.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/04/18 09:23:18
# @brief Definition of ForwardDISETTests class.

# # imports
import unittest

# # from DIRAC
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

# # SUT
from DIRAC.RequestManagementSystem.Agent.RequestOperations.ForwardDISET import ForwardDISET

########################################################################
class ForwardDISETTests(unittest.TestCase):
    """
    .. class:: ForwardDISETTests

    """

    def setUp(self):
        """test set up"""

        self.hiArgs = (
            ("RequestManagement/RequestManager", {"keepAliveLapse": 10, "timeout": 5}),
            "foo",
            (12345, {"Hi": "There!"}),
        )
        self.req = Request({"RequestName": "testRequest"})
        self.op = Operation({"Type": "ForwardDISET", "Arguments": DEncode.encode(self.hiArgs)})
        self.req += self.op

    def tearDown(self):
        """tear down"""
        del self.hiArgs
        del self.op
        del self.req

    def testCase(self):
        """ctor and functionality"""
        forwardDISET = None
        try:
            forwardDISET = ForwardDISET()
        except:
            pass
        self.assertEqual(isinstance(forwardDISET, ForwardDISET), True, "construction error")

        forwardDISET.setOperation(self.op)
        self.assertEqual(isinstance(forwardDISET.operation, Operation), True, "setOperation error")

        call = forwardDISET()
        # # should be failing right now
        self.assertEqual(call["OK"], False, "call failed")


if __name__ == "__main__":
    testLoader = unittest.TestLoader()
    forwardDISETTests = testLoader.loadTestsFromTestCase(ForwardDISETTests)
    suite = unittest.TestSuite([forwardDISETTests])
    unittest.TextTestRunner(verbosity=3).run(suite)



""" LineGraph represents line graphs both simple and stacked. It includes
    also cumulative graph functionality.

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.Graphs.PlotBase import PlotBase
from DIRAC.Core.Utilities.Graphs.GraphUtilities import (
    to_timestamp,
    PrettyDateLocator,
    PrettyDateFormatter,
    PrettyScalarFormatter,
)
from matplotlib.patches import Polygon
from matplotlib.dates import date2num
import datetime


class LineGraph(PlotBase):

    """
    The LineGraph class is a straightforward line graph; given a dictionary
    of values, it takes the keys as the independent variable and the values
    as the dependent variable.
    """

    def __init__(self, data, ax, prefs, *args, **kw):

        PlotBase.__init__(self, data, ax, prefs, *args, **kw)

    def draw(self):

        PlotBase.draw(self)
        self.x_formatter_cb(self.ax)

        if self.gdata.isEmpty():
            return None

        tmp_x = []
        tmp_y = []

        labels = self.gdata.getLabels()
        nKeys = self.gdata.getNumberOfKeys()
        tmp_b = []
        for n in range(nKeys):
            if "log_yaxis" in self.prefs:
                tmp_b.append(0.001)
            else:
                tmp_b.append(0.0)

        start_plot = 0
        end_plot = 0
        if "starttime" in self.prefs and "endtime" in self.prefs:
            start_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["starttime"])))
            end_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["endtime"])))

        self.polygons = []
        seq_b = [(self.gdata.max_num_key, 0.0), (self.gdata.min_num_key, 0.0)]
        zorder = 0.0
        labels = self.gdata.getLabels()
        labels.reverse()

        # If it is a simple plot, no labels are used
        # Evaluate the most appropriate color in this case
        if self.gdata.isSimplePlot():
            labels = [("SimplePlot", 0.0)]
            color = self.prefs.get("plot_color", "Default")
            if color.find("#") != -1:
                self.palette.setColor("SimplePlot", color)
            else:
                labels = [(color, 0.0)]

        for label, num in labels:

            color = self.palette.getColor(label)
            ind = 0
            tmp_x = []
            tmp_y = []
            plot_data = self.gdata.getPlotNumData(label)
            for key, value, error in plot_data:
                if value is None:
                    value = 0.0
                tmp_x.append(key)
                tmp_y.append(float(value) + tmp_b[ind])
                ind += 1
            seq_t = list(zip(tmp_x, tmp_y))
            seq = seq_t + seq_b
            poly = Polygon(seq, facecolor=color, fill=True, linewidth=0.2, zorder=zorder)
            self.ax.add_patch(poly)
            self.polygons.append(poly)
            tmp_b = list(tmp_y)
            zorder -= 0.1

        ymax = max(tmp_b)
        ymax *= 1.1
        ymin = min(min(tmp_b), 0.0)
        ymin *= 1.1
        if "log_yaxis" in self.prefs:
            ymin = 0.001
        xmax = max(tmp_x)
        if self.log_xaxis:
            xmin = 0.001
        else:
            xmin = 0

        ymin = self.prefs.get("ymin", ymin)
        ymax = self.prefs.get("ymax", ymax)
        xmin = self.prefs.get("xmin", xmin)
        xmax = self.prefs.get("xmax", xmax)

        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        if self.gdata.key_type == "time":
            if start_plot and end_plot:
                self.ax.set_xlim(xmin=start_plot, xmax=end_plot)
            else:
                self.ax.set_xlim(xmin=min(tmp_x), xmax=max(tmp_x))

    def x_formatter_cb(self, ax):
        if self.gdata.key_type == "string":
            smap = self.gdata.getStringMap()
            reverse_smap = {}
            for key, val in smap.items():
                reverse_smap[val] = key
            ticks = smap.values()
            ticks.sort()
            ax.set_xticks([i + 0.5 for i in ticks])
            ax.set_xticklabels([reverse_smap[i] for i in ticks])
            ax.grid(False)
            if self.log_xaxis:
                xmin = 0.001
            else:
                xmin = 0
            ax.set_xlim(xmin=xmin, xmax=len(ticks))
        elif self.gdata.key_type == "time":
            dl = PrettyDateLocator()
            df = PrettyDateFormatter(dl)
            ax.xaxis.set_major_locator(dl)
            ax.xaxis.set_major_formatter(df)
            ax.xaxis.set_clip_on(False)
            sf = PrettyScalarFormatter()
            ax.yaxis.set_major_formatter(sf)

        else:
            return None



""" DIRAC Workload Management System Client class encapsulates all the
    methods necessary to communicate with the Workload Management System
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from six import StringIO
import time

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities import File
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.DErrno import EWMSJDL, EWMSSUBM
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient import SandboxStoreClient
from DIRAC.WorkloadManagementSystem.Utilities.ParametricJob import getParameterVectorLength

__RCSID__ = "$Id$"


class WMSClient(object):
    """Class exposing the following jobs methods:

    submit
    kill
    delete
    remove
    reschedule
    reset
    """

    def __init__(
        self,
        jobManagerClient=None,
        sbRPCClient=None,
        sbTransferClient=None,
        useCertificates=False,
        timeout=600,
        delegatedDN=None,
        delegatedGroup=None,
    ):
        """WMS Client constructor

        Here we also initialize the needed clients and connections
        """

        self.useCertificates = useCertificates
        self.delegatedDN = delegatedDN
        self.delegatedGroup = delegatedGroup
        self.timeout = timeout
        self._jobManager = jobManagerClient
        self.operationsHelper = Operations()
        self.sandboxClient = None
        if sbRPCClient and sbTransferClient:
            self.sandboxClient = SandboxStoreClient(
                rpcClient=sbRPCClient, transferClient=sbTransferClient, useCertificates=useCertificates
            )

    @property
    def jobManager(self):
        if not self._jobManager:
            self._jobManager = JobManagerClient(
                useCertificates=self.useCertificates,
                delegatedDN=self.delegatedDN,
                delegatedGroup=self.delegatedGroup,
                timeout=self.timeout,
            )

        return self._jobManager

    def __getInputSandboxEntries(self, classAdJob):
        if classAdJob.lookupAttribute("InputSandbox"):
            inputSandbox = classAdJob.get_expression("InputSandbox")
            inputSandbox = inputSandbox.replace('","', "\n")
            inputSandbox = inputSandbox.replace("{", "")
            inputSandbox = inputSandbox.replace("}", "")
            inputSandbox = inputSandbox.replace('"', "")
            inputSandbox = inputSandbox.replace(",", "")
            inputSandbox = inputSandbox.split()
        else:
            inputSandbox = []

        return inputSandbox

    def __uploadInputSandbox(self, classAdJob, jobDescriptionObject=None):
        """Checks the validity of the job Input Sandbox.
        The function returns the list of Input Sandbox files.
        The total volume of the input sandbox is evaluated
        """
        inputSandbox = self.__getInputSandboxEntries(classAdJob)

        realFiles = []
        badFiles = []
        diskFiles = []

        for isFile in inputSandbox:
            if not isFile.startswith(("lfn:", "LFN:", "SB:", "%s", "%(")):
                realFiles.append(isFile)

        stringIOFiles = []
        stringIOFilesSize = 0
        if jobDescriptionObject is not None:
            if isinstance(jobDescriptionObject, StringIO):
                stringIOFiles = [jobDescriptionObject]
                stringIOFilesSize = len(jobDescriptionObject.getvalue())
                gLogger.debug("Size of the stringIOFiles: " + str(stringIOFilesSize))
            else:
                return S_ERROR(EWMSJDL, "jobDescriptionObject is not a StringIO object")

        # Check real files
        for isFile in realFiles:
            if not os.path.exists(isFile):  # we are passing in real files, we expect them to be on disk
                badFiles.append(isFile)
                gLogger.warn("inputSandbox file/directory " + isFile + " not found. Keep looking for the others")
                continue
            diskFiles.append(isFile)

        diskFilesSize = File.getGlobbedTotalSize(diskFiles)
        gLogger.debug("Size of the diskFiles: " + str(diskFilesSize))
        totalSize = diskFilesSize + stringIOFilesSize
        gLogger.verbose("Total size of the inputSandbox: " + str(totalSize))

        okFiles = stringIOFiles + diskFiles
        if badFiles:
            result = S_ERROR(EWMSJDL, "Input Sandbox is not valid")
            result["BadFile"] = badFiles
            result["TotalSize"] = totalSize
            return result

        if okFiles:
            if not self.sandboxClient:
                self.sandboxClient = SandboxStoreClient(
                    useCertificates=self.useCertificates,
                    delegatedDN=self.delegatedDN,
                    delegatedGroup=self.delegatedGroup,
                )
            result = self.sandboxClient.uploadFilesAsSandbox(okFiles)
            if not result["OK"]:
                return result
            inputSandbox.append(result["Value"])
            classAdJob.insertAttributeVectorString("InputSandbox", inputSandbox)

        return S_OK()

    def submitJob(self, jdl, jobDescriptionObject=None):
        """Submit one job specified by its JDL to WMS.

        The JDL may actually be the desciption of a parametric job,
        resulting in multiple DIRAC jobs submitted to the DIRAC WMS
        """

        if os.path.exists(jdl):
            with open(jdl, "r") as fic:
                jdlString = fic.read()
        else:
            # If file JDL does not exist, assume that the JDL is passed as a string
            jdlString = jdl

        jdlString = jdlString.strip()

        gLogger.debug("Submitting JDL", jdlString)
        # Strip of comments in the jdl string
        newJdlList = []
        for line in jdlString.split("\n"):
            if not line.strip().startswith("#"):
                newJdlList.append(line)
        jdlString = "\n".join(newJdlList)

        # Check the validity of the input JDL
        if jdlString.find("[") != 0:
            jdlString = "[%s]" % jdlString
        classAdJob = ClassAd(jdlString)
        if not classAdJob.isOK():
            return S_ERROR(EWMSJDL, "Invalid job JDL")

        # Check the size and the contents of the input sandbox
        result = self.__uploadInputSandbox(classAdJob, jobDescriptionObject)
        if not result["OK"]:
            return result

        # Submit the job now and get the new job ID
        result = getParameterVectorLength(classAdJob)
        if not result["OK"]:
            return result
        nJobs = result["Value"]
        result = self.jobManager.submitJob(classAdJob.asJDL())

        if nJobs:
            gLogger.debug("Applying transactional job submission")
            # The server applies transactional bulk submission, we should confirm the jobs
            if result["OK"]:
                jobIDList = result["Value"]
                if len(jobIDList) == nJobs:
                    # Confirm the submitted jobs
                    confirmed = False
                    for _attempt in range(3):
                        result = self.jobManager.confirmBulkSubmission(jobIDList)
                        if result["OK"]:
                            confirmed = True
                            break
                        time.sleep(1)
                    if not confirmed:
                        # The bulk submission failed, try to remove the created jobs
                        resultDelete = self.jobManager.removeJob(jobIDList)
                        error = "Job submission failed to confirm bulk transaction"
                        if not resultDelete["OK"]:
                            error += "; removal of created jobs failed"
                        return S_ERROR(EWMSSUBM, error)
                else:
                    return S_ERROR(EWMSSUBM, "The number of submitted jobs does not match job description")

        if result.get("requireProxyUpload"):
            gLogger.warn("Need to upload the proxy")

        return result

    def killJob(self, jobID):
        """Kill running job.
        jobID can be an integer representing a single DIRAC job ID or a list of IDs
        """
        return self.jobManager.killJob(jobID)

    def deleteJob(self, jobID):
        """Delete job(s) (set their status to DELETED) from the WMS Job database.
        jobID can be an integer representing a single DIRAC job ID or a list of IDs
        """
        return self.jobManager.deleteJob(jobID)

    def removeJob(self, jobID):
        """Fully remove job(s) from the WMS Job database.
        jobID can be an integer representing a single DIRAC job ID or a list of IDs
        """
        return self.jobManager.removeJob(jobID)

    def rescheduleJob(self, jobID):
        """Reschedule job(s) in WMS Job database.
        jobID can be an integer representing a single DIRAC job ID or a list of IDs
        """
        return self.jobManager.rescheduleJob(jobID)

    def resetJob(self, jobID):
        """Reset job(s) in WMS Job database.
        jobID can be an integer representing a single DIRAC job ID or a list of IDs
        """
        return self.jobManager.resetJob(jobID)



"""
Usage of ThreadPool

ThreadPool creates a pool of worker threads to process a queue of tasks
much like the producers/consumers paradigm. Users just need to fill the queue
with tasks to be executed and worker threads will execute them

To start working with the ThreadPool first it has to be instanced::

    threadPool = ThreadPool( minThreads, maxThreads, maxQueuedRequests )

    minThreads        -> at all times no less than <minThreads> workers will be alive
    maxThreads        -> at all times no more than <maxThreads> workers will be alive
    maxQueuedRequests -> No more than <maxQueuedRequests> can be waiting to be executed
                         If another request is added to the ThreadPool, the thread will
                         lock until another request is taken out of the queue.

The ThreadPool will automatically increase and decrease the pool of workers as needed

To add requests to the queue::

     threadPool.generateJobAndQueueIt( <functionToExecute>,
                                       args = ( arg1, arg2, ... ),
                                       oCallback = <resultCallbackFunction> )

or::

     request = ThreadedJob( <functionToExecute>,
                            args = ( arg1, arg2, ... )
                            oCallback = <resultCallbackFunction> )
     threadPool.queueJob( request )

The result callback and the parameters are optional arguments.
Once the requests have been added to the pool. They will be executed as soon as possible.
Worker threads automatically return the return value of the requests. To run the result callback
functions execute::

   threadPool.generateJobAndQueueIt( <functionToExecute>,
                                     args = ( arg1, arg2, ... ),
                                     oCallback = <resultCallbackFunction> )

or::

   request = ThreadedJob( <functionToExecute>,
                          args = ( arg1, arg2, ... )
                          oCallback = <resultCallbackFunction> )
   threadPool.queueJob( request )

The result callback and the parameters are optional arguments.
Once the requests have been added to the pool. They will be executed as soon as possible.
Worker threads automatically return the return value of the requests. To run the result callback
functions execute::

   threadPool.processRequests()

This method will process the existing return values of the requests. Even if the requests do not return
anything this method (or any process result method) has to be called to clean the result queues.

To wait until all the requests are finished and process their result call::

   threadPool.processAllRequests()

This function will block until all requests are finished and their result values have been processed.

It is also possible to set the threadPool in auto processing results mode. It'll process the results as
soon as the requests have finished. To enable this mode call::

   threadPool.daemonize()

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import time
import sys
from six.moves import queue as Queue
import threading

try:
    from DIRAC.FrameworkSystem.Client.Logger import gLogger
except Exception:
    gLogger = False
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR


class WorkingThread(threading.Thread):
    def __init__(self, oPendingQueue, oResultsQueue, **kwargs):
        threading.Thread.__init__(self, **kwargs)
        self.setDaemon(1)
        self.__pendingQueue = oPendingQueue
        self.__resultsQueue = oResultsQueue
        self.__threadAlive = True
        self.__working = False
        self.start()

    def isWorking(self):
        return self.__working

    def kill(self):
        self.__threadAlive = False

    def run(self):
        while self.__threadAlive:
            oJob = self.__pendingQueue.get(block=True)
            if not self.__threadAlive:
                self.__pendingQueue.put(oJob)
                break
            self.__working = True
            oJob.process()
            self.__working = False
            if oJob.hasCallback():
                self.__resultsQueue.put(oJob, block=True)


class ThreadedJob(object):
    def __init__(self, oCallable, args=None, kwargs=None, sTJId=None, oCallback=None, oExceptionCallback=None):
        self.__jobFunction = oCallable
        self.__jobArgs = args or []
        self.__jobKwArgs = kwargs or {}
        self.__tjID = sTJId
        self.__resultCallback = oCallback
        self.__exceptionCallback = oExceptionCallback
        self.__done = False
        self.__exceptionRaised = False
        self.__jobResult = None
        self.__jobException = None

    def __showException(self, threadedJob, exceptionInfo):
        if gLogger:
            gLogger.exception("Exception in thread", lExcInfo=exceptionInfo)

    def jobId(self):
        return self.__tjID

    def hasCallback(self):
        return self.__resultCallback or self.__exceptionCallback

    def exceptionRaised(self):
        return self.__exceptionRaised

    def doExceptionCallback(self):
        if self.__done and self.__exceptionRaised and self.__exceptionCallback:
            self.__exceptionCallback(self, self.__jobException)

    def doCallback(self):
        if self.__done and not self.__exceptionRaised and self.__resultCallback:
            self.__resultCallback(self, self.__jobResult)

    def process(self):
        self.__done = True
        try:
            self.__jobResult = self.__jobFunction(*self.__jobArgs, **self.__jobKwArgs)
        except Exception as lException:
            self.__exceptionRaised = True
            if not self.__exceptionCallback:
                if gLogger:
                    gLogger.exception("Exception in thread", lException=lException)
            else:
                self.__jobException = sys.exc_info()


class ThreadPool(threading.Thread):
    def __init__(self, iMinThreads, iMaxThreads=0, iMaxQueuedRequests=0, strictLimits=True):
        threading.Thread.__init__(self)
        if iMinThreads < 1:
            self.__minThreads = 1
        else:
            self.__minThreads = iMinThreads
        if iMaxThreads < self.__minThreads:
            self.__maxThreads = self.__minThreads
        else:
            self.__maxThreads = iMaxThreads
        self.__strictLimits = strictLimits
        self.__pendingQueue = Queue.Queue(iMaxQueuedRequests)
        self.__resultsQueue = Queue.Queue(iMaxQueuedRequests + iMaxThreads)
        self.__workingThreadsList = []
        self.__spawnNeededWorkingThreads()

    def getMaxThreads(self):
        return self.__maxThreads

    def getMinThreads(self):
        return self.__minThreads

    def numWorkingThreads(self):
        return self.__countWorkingThreads()

    def numWaitingThreads(self):
        return self.__countWaitingThreads()

    def __spawnWorkingThread(self):
        self.__workingThreadsList.append(WorkingThread(self.__pendingQueue, self.__resultsQueue))

    def __killWorkingThread(self):
        if self.__strictLimits:
            for i in range(len(self.__workingThreadsList)):
                wT = self.__workingThreadsList[i]
                if not wT.isWorking():
                    wT.kill()
                    del self.__workingThreadsList[i]
                    break
        else:
            self.__workingThreadsList[0].kill()
            del self.__workingThreadsList[0]

    def __countWaitingThreads(self):
        iWaitingThreads = 0
        for oWT in self.__workingThreadsList:
            if not oWT.isWorking():
                iWaitingThreads += 1
        return iWaitingThreads

    def __countWorkingThreads(self):
        iWorkingThreads = 0
        for oWT in self.__workingThreadsList:
            if oWT.isWorking():
                iWorkingThreads += 1
        return iWorkingThreads

    def __spawnNeededWorkingThreads(self):
        while len(self.__workingThreadsList) < self.__minThreads:
            self.__spawnWorkingThread()
        while self.__countWaitingThreads() == 0 and len(self.__workingThreadsList) < self.__maxThreads:
            self.__spawnWorkingThread()

    def __killExceedingWorkingThreads(self):
        threadsToKill = len(self.__workingThreadsList) - self.__maxThreads
        for _ in range(max(threadsToKill, 0)):
            self.__killWorkingThread()
        threadsToKill = self.__countWaitingThreads() - self.__minThreads
        for _ in range(max(threadsToKill, 0)):
            self.__killWorkingThread()

    def queueJob(self, oTJob, blocking=True):
        if not isinstance(oTJob, ThreadedJob):
            raise TypeError("Jobs added to the thread pool must be ThreadedJob instances")
        try:
            self.__pendingQueue.put(oTJob, block=blocking)
        except Queue.Full:
            return S_ERROR("Queue is full")
        return S_OK()

    def generateJobAndQueueIt(
        self, oCallable, args=None, kwargs=None, sTJId=None, oCallback=None, oExceptionCallback=None, blocking=True
    ):
        oTJ = ThreadedJob(oCallable, args, kwargs, sTJId, oCallback, oExceptionCallback)
        return self.queueJob(oTJ, blocking)

    def pendingJobs(self):
        return self.__pendingQueue.qsize()

    def isFull(self):
        return self.__pendingQueue.full()

    def isWorking(self):
        return not self.__pendingQueue.empty() or self.__countWorkingThreads()

    def processResults(self):
        iProcessed = 0
        while True:
            self.__spawnNeededWorkingThreads()
            if self.__resultsQueue.empty():
                self.__killExceedingWorkingThreads()
                break
            oJob = self.__resultsQueue.get()
            oJob.doExceptionCallback()
            oJob.doCallback()
            iProcessed += 1
            self.__killExceedingWorkingThreads()
        return iProcessed

    def processAllResults(self):
        while not self.__pendingQueue.empty() or self.__countWorkingThreads():
            self.processResults()
            time.sleep(0.1)
        self.processResults()

    def daemonize(self):
        self.setDaemon(1)
        self.start()

    # This is the ThreadPool threaded function. YOU ARE NOT SUPPOSED TO CALL THIS FUNCTION!!!
    def run(self):
        while True:
            self.processResults()
            time.sleep(1)


gThreadPool = False


def getGlobalThreadPool():
    global gThreadPool
    if not gThreadPool:
        gThreadPool = ThreadPool(1, 500)
        gThreadPool.daemonize()
    return gThreadPool


if __name__ == "__main__":
    import random

    def doSomething(iNumber):
        time.sleep(random.randint(1, 5))
        fResult = random.random() * iNumber
        if fResult > 3:
            raise Exception("TEST EXCEPTION")
        return fResult

    def showResult(oTJ, fResult):
        print("Result %s from %s" % (fResult, oTJ))

    def showException(oTJ, exc_info):
        print("Exception %s from %s" % (exc_info[1], oTJ))

    OTP = ThreadPool(5, 10)

    def generateWork(iWorkUnits):
        for iNumber in [random.randint(1, 20) for _ in range(iWorkUnits)]:
            oTJ = ThreadedJob(doSomething, args=(iNumber,), oCallback=showResult, oExceptionCallback=showException)
            OTP.queueJob(oTJ)

    print("MaxThreads =", OTP.getMaxThreads())
    print("MinThreads =", OTP.getMinThreads())

    generateWork(30)
    while True:
        time.sleep(1)
        gIResult = OTP.processResults()
        gINew = gIResult + random.randint(-3, 2)
        print("Processed %s, generating %s.." % (gIResult, gINew))
        generateWork(gINew)
        print("Threads %s" % OTP.numWorkingThreads(), OTP.pendingJobs())



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter
from DIRAC.AccountingSystem.Client.Types.PilotSubmission import PilotSubmission

__RCSID__ = "$Id$"


class PilotSubmissionPlotter(BaseReporter):
    """
    Plotter for the Pilot Submission Accounting
    """

    _typeName = "PilotSubmission"
    _typeKeyFields = [dF[0] for dF in PilotSubmission().definitionKeyFields]

    def _reportSubmission(self, reportRequest):
        """
        Get data for Submission plot

        :param dict reportRequest: Condition to select data
        """

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "NumTotal"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, _ = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, _, unit = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK({"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unit})

    def _plotSubmission(self, reportRequest, plotInfo, filename):
        """
        Make 1 dimensional pilotSubmission plot

        :param dict reportRequest: Condition to select data
        :param dict plotInfo: Data for plot
        :param str  filename: File name
        """

        metadata = {
            "title": "Number of Submission by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateTimedStackedBarPlot(filename, plotInfo["graphDataDict"], metadata)

    _reportSubmissionEfficiencyName = "Submission efficiency"

    def _reportSubmissionEfficiency(self, reportRequest):
        """
        Get data for Submission Efficiency plot

        :param dict reportRequest: Condition to select data
        """

        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "NumSucceeded", "NumTotal"],
        )

        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {
                "checkNone": True,
                "convertToGranularity": "sum",
                "calculateProportionalGauges": False,
                "consolidationFunction": self._efficiencyConsolidation,
            },
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        if len(dataDict) > 1:
            # Get the total for the plot
            selectFields = (
                "'Total', %s, %s, SUM(%s),SUM(%s)",
                ["startTime", "bucketLength", "NumSucceeded", "NumTotal"],
            )

            retVal = self._getTimedData(
                reportRequest["startTime"],
                reportRequest["endTime"],
                selectFields,
                reportRequest["condDict"],
                reportRequest["groupingFields"],
                {
                    "scheckNone": True,
                    "convertToGranularity": "sum",
                    "calculateProportionalGauges": False,
                    "consolidationFunction": self._efficiencyConsolidation,
                },
            )
            if not retVal["OK"]:
                return retVal
            totalDict = retVal["Value"][0]
            self.stripDataField(totalDict, 0)
            for key in totalDict:
                dataDict[key] = totalDict[key]
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotSubmissionEfficiency(self, reportRequest, plotInfo, filename):
        """
        Make 2 dimensional pilotSubmission efficiency plot

        :param dict reportRequest: Condition to select data
        :param dict plotInfo: Data for plot.
        :param str  filename: File name
        """

        metadata = {
            "title": "Pilot Submission efficiency by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }

        return self._generateQualityPlot(filename, plotInfo["data"], metadata)



""" DIRAC FileCatalog base class for mix-in classes to manage users and groups
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import threading

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.Core.Security import Properties


class UserAndGroupManagerBase(object):
    def _refreshGroups(self):
        """Refresh the group cache"""
        return S_ERROR("To be implemented on derived class")

    def _refreshUsers(self):
        """Refresh the user cache"""
        return S_ERROR("To be implemented on derived class")

    def __init__(self, database=None):
        self.db = database
        self.lock = threading.Lock()
        self._refreshUsers()
        self._refreshGroups()

    def setDatabase(self, database):
        self.db = database

    def getUserAndGroupRight(self, credDict):
        """Evaluate rights for user and group operations"""
        if Properties.FC_MANAGEMENT in credDict["properties"]:
            return S_OK(True)
        return S_OK(False)



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-monitor-jobs
# Author :  Daniela Remenska
########################################################################
"""
Report the summary of the staging progress of jobs

Example:
  $ dirac-stager-monitor-jobs.py 5688643 5688644

  JobID               : 5688643
  Status              : Offline
  SubmitTime          : 2013-06-10 15:21:03
  CompleteTime        : None
  Staging files for this job:
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00003705_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00003705_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00001918_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00001918_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00002347_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00002347_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00003701_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00003701_1.sdst
      Status  : Offline
      Reason  : None
  ----------------------
  JobID               : 5688644
  Status              : Offline
  SubmitTime          : 2013-06-10 15:21:07
  CompleteTime        : None
  Staging files for this job:
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00005873_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00005873_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00004468_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00004468_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00000309_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00000309_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00005911_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00005911_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
      LFN     : /lhcb/LHCb/Collision10/SDST/01/0000/01_00003296_1.sdst
      SE      : IN2P3-RDST
      PFN     : srm://ccsrm.in2p3.fr/pnfs/in2p3.fr/data/lhcb/LHCb/Collision10/SDST/01/0000/01_00003296_1.sdst
      Status  : Offline
      Reason  : None
      --------------------
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    Script.parseCommandLine(ignoreErrors=False)

    args = Script.getPositionalArgs()

    if len(args) < 1:
        Script.showHelp()

    from DIRAC import exit as DIRACExit, gLogger

    try:
        jobIDs = [int(arg) for arg in args]
    except Exception:
        gLogger.fatal("DIRAC Job IDs must be integers")
        DIRACExit(2)

    from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

    client = StorageManagerClient()

    outStr = "\n"
    for jobID in jobIDs:
        res = client.getTaskSummary(jobID)
        if not res["OK"]:
            gLogger.error(res["Message"])
            continue
        if not res["Value"]:
            gLogger.notice("No info for job %s, probably gone from the stager..." % jobID)
            continue
        taskInfo = res["Value"]["TaskInfo"]
        replicaInfo = res["Value"]["ReplicaInfo"]
        outStr = "%s: %s" % ("JobID".ljust(20), jobID)
        outStr += "\n%s: %s" % ("Status".ljust(20), taskInfo[str(jobID)]["Status"])
        outStr += "\n%s: %s" % ("SubmitTime".ljust(20), taskInfo[str(jobID)]["SubmitTime"])
        outStr += "\n%s: %s" % ("CompleteTime".ljust(20), taskInfo[str(jobID)]["CompleteTime"])
        outStr += "\nStaging files for this job:"
        if not res["Value"]["ReplicaInfo"]:
            gLogger.notice("No info on files for the job = %s, that is odd" % jobID)
            continue
        else:
            for lfn, metadata in replicaInfo.items():
                outStr += "\n\t--------------------"
                outStr += "\n\t%s: %s" % ("LFN".ljust(8), lfn.ljust(100))
                outStr += "\n\t%s: %s" % ("SE".ljust(8), metadata["StorageElement"].ljust(100))
                outStr += "\n\t%s: %s" % ("PFN".ljust(8), str(metadata["PFN"]).ljust(100))
                outStr += "\n\t%s: %s" % ("Status".ljust(8), metadata["Status"].ljust(100))
                outStr += "\n\t%s: %s" % ("Reason".ljust(8), str(metadata["Reason"]).ljust(100))
                outStr += "\n%s: %s" % ("LastUpdate".ljust(8), str(metadata["LastUpdate"]).ljust(100))
            outStr += "\n----------------------"
        gLogger.notice(outStr)
    DIRACExit(0)


if __name__ == "__main__":
    main()



""" DIRACAccountingCommand

  The DIRACAccountingCommand class is a command class to
  interrogate the DIRAC Accounting.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# FIXME: NOT Usable ATM
# missing doNew, doCache, doMaster

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.AccountingSystem.Client.ReportsClient import ReportsClient
from DIRAC.ResourceStatusSystem.Command.Command import Command
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient

################################################################################
################################################################################


class DIRACAccountingCommand(Command):
    def __init__(self, args=None, clients=None):

        super(DIRACAccountingCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns jobs accounting info for sites in the last 24h

        Args:
           - args[0]: string - should be a ValidElement

           - args[1]: string - should be the name of the ValidElement

           - args[2]: string - should be 'Job' or 'Pilot' or 'DataOperation'
             or 'WMSHistory' (??)

           - args[3]: string - should be the plot to generate (e.g. CPUEfficiency)

           - args[4]: dictionary - e.g. {'Format': 'LastHours', 'hours': 24}

           - args[5]: string - should be the grouping

           - args[6]: dictionary - optional conditions
        """

        granularity = self.args[0]
        name = self.args[1]
        accounting = self.args[2]
        plot = self.args[3]
        period = self.args[4]
        grouping = self.args[5]

        if period["Format"] == "LastHours":
            fromT = datetime.utcnow() - timedelta(hours=period["hours"])
            toT = datetime.utcnow()
        elif period["Format"] == "Periods":
            # TODO
            pass

        if self.args[6] is not None:
            conditions = self.args[6]
        else:
            conditions = {}
            if accounting == "Job" or accounting == "Pilot":
                if granularity == "Resource":
                    conditions["GridCE"] = [name]
                elif granularity == "Service":
                    conditions["Site"] = [name.split("@").pop()]
                elif granularity == "Site":
                    conditions["Site"] = [name]
                else:
                    return S_ERROR("%s is not a valid granularity" % granularity)
            elif accounting == "DataOperation":
                conditions["Destination"] = [name]

        return self.rClient.getReport(accounting, plot, fromT, toT, conditions, grouping)


################################################################################
################################################################################


class TransferQualityCommand(Command):
    def __init__(self, args=None, clients=None):

        super(TransferQualityCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Return getQuality from DIRAC's accounting ReportsClient

        `args`: a tuple
          - args[0]: string: should be a ValidElement

          - args[1]: string should be the name of the ValidElement

          - args[2]: optional dateTime object: a "from" date

          - args[3]: optional dateTime object: a "to" date

        :returns:
          {'Result': None | a float between 0.0 and 100.0}
        """

        if "fromDate" not in self.args:
            fromDate = datetime.utcnow() - timedelta(hours=2)
        else:
            fromDate = self.args["fromDate"]

        if "toDate" not in self.args:
            toDate = datetime.utcnow()
        else:
            toDate = self.args["toDate"]

        if "name" not in self.args:
            return S_ERROR("name not specified")
        name = self.args["name"]

        results = self.rClient.getReport(
            "DataOperation",
            "Quality",
            fromDate,
            toDate,
            {"OperationType": "putAndRegister", "Destination": [name]},
            "Channel",
        )

        if not results["OK"]:
            return results

        pr_q_d = {channel: strToIntDict(value) for channel, value in results["Value"]["data"].items()}

        # FIXME: WHAT the hell is this doing ?
        values = []
        if len(pr_q_d) == 1:

            for k in pr_q_d:
                for n in pr_q_d[k].values():
                    values.append(n)
            res = sum(values) / len(values)  # FIXME: should convert to int?

        else:
            for n in pr_q_d["Total"].values():
                values.append(n)
            res = sum(values) / len(values)  # FIXME: should convert to int?

        return S_OK(res)


################################################################################
################################################################################
#
# class TransferQualityCached_Command(Command):
#
#  __APIs__ = [ 'ResourceManagementClient' ]
#
#  def doCommand(self):
#    """
#    Returns transfer quality as it is cached
#
#    :attr:`args`:
#       - args[0]: string: should be a ValidElement
#
#       - args[1]: string should be the name of the ValidElement
#
#    :returns:
#      {'Result': None | a float between 0.0 and 100.0}
#    """
#
#    super(TransferQualityCached_Command, self).doCommand()
#    self.apis = initAPIs( self.__APIs__, self.apis )
#
#    name = self.args[1]
#
#    try:
#      res = self.apis[ 'ResourceManagementClient' ].getCachedResult(name, 'TransferQualityEverySEs', 'TQ', 'NULL')
#      if res == []:
#        return {'Result':None}
#    except:
#      gLogger.exception("Exception when calling ResourceManagementClient for %s" %(name))
#      return {'Result':'Unknown'}
#
#    return {'Result':float(res[0])}
#
#  doCommand.__doc__ = Command.doCommand.__doc__ + doCommand.__doc__
#
################################################################################
################################################################################


class CachedPlotCommand(Command):
    def __init__(self, args=None, clients=None):

        super(CachedPlotCommand, self).__init__(args, clients)

        if "ResourceManagementClient" in self.apis:
            self.rmClient = self.apis["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def doCommand(self):
        """
        Returns transfer quality plot as it is cached in the accounting cache.

        :attr:`args`:
           - args[0]: string - should be a ValidElement

           - args[1]: string - should be the name of the ValidElement

           - args[2]: string - should be the plot type

           - args[3]: string - should be the plot name

        :returns:
          a plot
        """

        if "element" not in self.args:
            return S_ERROR("element no specified")
        element = self.args["element"]

        if "name" not in self.args:
            return S_ERROR("Name no specified")
        name = self.args["name"]

        if "plotType" not in self.args:
            return S_ERROR("plotType no specified")
        plotType = self.args["plotType"]

        if "plotName" not in self.args:
            return S_ERROR("plotName no specified")
        plotName = self.args["plotName"]

        # FIXME: we have no any longer Service granularity !
        if element == "Service":
            name = name.split("@")[1]

        meta = {"columns": "Result"}

        results = self.rmClient.selectAccountingCache(name=name, plotType=plotType, plotName=plotName, meta=meta)

        if not results["OK"]:
            return results
        results = results["Value"]

        if results == []:
            results = {"data": {}, "granularity": 900}
        else:
            # FIXME: WTH is an eval doing here !!!!
            results = eval(results[0])

        return results


################################################################################
################################################################################


class TransferQualityFromCachedPlotCommand(Command):
    def __init__(self, args=None, clients=None):

        super(TransferQualityFromCachedPlotCommand, self).__init__(args, clients)

        if "ResourceManagementClient" in self.apis:
            self.rmClient = self.apis["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def doCommand(self):
        """
        Returns transfer quality from the plot cached in the accounting cache.

        :attr:`args`:
           - args[0]: string: should be a ValidElement

           - args[1]: string should be the name of the ValidElement

        :returns:
          {'Result': None | a float between 0.0 and 100.0}
        """

        if "name" not in self.args:
            return S_ERROR("Name no specified")
        name = self.args["name"]

        if "plotType" not in self.args:
            return S_ERROR("plotType no specified")
        plotType = self.args["plotType"]

        if "plotName" not in self.args:
            return S_ERROR("plotName no specified")
        plotName = self.args["plotName"]

        meta = {"columns": "Result"}

        results = self.rmClient.selectAccountingCache(name=name, plotType=plotType, plotName=plotName, meta=meta)

        if not results["OK"]:
            return results
        results = results["Value"]

        if results == []:
            results = None
        else:
            # FIXME: remove the eval from here !!
            results = eval(results[0][0])

            num, den = 0, 0

            se = list(results["data"])[0]

            num = num + len(results["data"][se])
            den = den + sum(results["data"][se].values())
            meanQuality = den / num

            results = meanQuality

        return S_OK(results)


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" Utility Class for threaded agents (e.g. TransformationAgent)
    Mostly for logging
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
from DIRAC import gLogger

__RCSID__ = "$Id$"

AGENT_NAME = ""


class TransformationAgentsUtilities(object):
    """logging utilities for threaded TS agents"""

    def __init__(self):
        """c'tor"""
        self.debug = False

    def __prefixForLogging(self, transID, method, reftime):
        """get the thread number"""
        if reftime is not None:
            method += " (%.1f seconds)" % (time.time() - reftime)
        try:
            return "[%s] " % transID + AGENT_NAME + "." + method
        except NameError:
            return ""

    def _logVerbose(self, message, param="", method="execute", transID="None", reftime=None):
        """verbose"""
        if self.debug:
            gLogger.getSubLogger("(V) " + self.__prefixForLogging(transID, method, reftime)).info(message, param)
        else:
            gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).verbose(message, param)

    def _logDebug(self, message, param="", method="execute", transID="None", reftime=None):
        """debug"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).debug(message, param)

    def _logInfo(self, message, param="", method="execute", transID="None", reftime=None):
        """info"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).info(message, param)

    def _logWarn(self, message, param="", method="execute", transID="None", reftime=None):
        """warn"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).warn(message, param)

    def _logError(self, message, param="", method="execute", transID="None", reftime=None):
        """error"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).error(message, param)

    def _logException(self, message, param="", lException=False, method="execute", transID="None", reftime=None):
        """exception"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).exception(message, param, lException)

    def _logFatal(self, message, param="", method="execute", transID="None", reftime=None):
        """error"""
        gLogger.getSubLogger(self.__prefixForLogging(transID, method, reftime)).fatal(message, param)

    def _transTaskName(self, transID, taskID):  # pylint: disable=no-self-use
        """Construct the task name from the transformation and task ID"""
        return str(transID).zfill(8) + "_" + str(taskID).zfill(8)

    def _parseTaskName(self, taskName):  # pylint: disable=no-self-use
        """Split a task name into transformation and taskID"""
        try:
            return (int(x) for x in taskName.split("_"))
        except ValueError:
            return (0, 0)



""" This is a test of the creation of the json dump file
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import os
from diraccfg import CFG

from DIRAC.WorkloadManagementSystem.Utilities.PilotCStoJSONSynchronizer import PilotCStoJSONSynchronizer
from DIRAC.ConfigurationSystem.private.ConfigurationClient import ConfigurationClient
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

# pylint: disable=protected-access


class PilotCStoJSONSynchronizerTestCase(unittest.TestCase):
    """Base class for the PilotCStoJSONSynchronizer test cases"""

    def setUp(self):
        # Creating test configuration file
        self.clearCFG()

        self.testCfgFileName = "test.cfg"
        cfgContent = """
    DIRAC
    {
      Setup=TestSetup
      Setups
      {
        TestSetup
        {
          WorkloadManagement=MyWM
        }
      }
    }
    Systems
    {
      WorkloadManagement
      {
        MyWM
        {
          URLs
          {
            Service1 = dips://server1:1234/WorkloadManagement/Service1
            Service2 = dips://$MAINSERVERS$:5678/WorkloadManagement/Service2
          }
          FailoverURLs
          {
            Service2 = dips://failover1:5678/WorkloadManagement/Service2
          }
        }
      }
    }
    Operations
    {
      Defaults
      {
        Pilot
        {
          Project = LHCb
          GenericPilotDN = /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=doe/CN=111213/CN=Joe Doe
          GenericPilotGroup = xxx_pilot
        }

        MainServers = gw1, gw2
      }
    }
    Registry
    {
      Users
      {
        ttester
        {
          DN = /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=ttester/CN=696969/CN=Thomas Tester
          CA = /DC=ch/DC=cern/CN=CERN Grid Certification Authority
          Email = thomas.tester@cern.ch
        }
        franekbolek
        {
          DN = /DC=ch/DC=voodo/OU=Organic Units/OU=Users/CN=franekbolek/CN=111122/CN=Franek Bolek
          CA = /DC=ch/DC=voodo/CN=Voodo Grid Certification Authority
          Email = franek.bolek@voodo.pl
        }
      }
      Groups
      {
        lhcb_pilot
        {
          #@@-host - /DC=ch/DC=voodo/OU=computers/CN=brabra.voodo.pl
          Users = franekbolek
          Users += ttester
          Properties = GenericPilot
          Properties += LimitedDelegation
          VOMSRole = /lhcb/Role=pilot
          #@@-ggg@diracAdmin - 2015-07-07 13:40:55
          VO = lhcb
        }
      }
    }
    Resources
    {
      Sites
      {
        Tests
        {
           Tests.Testing.tst
           {
             CEs
             {
                test1.Testing.tst
                {
                  CEType = Tester
                }
             }
           }
        }
      }
    }
    """
        with open(self.testCfgFileName, "w") as f:
            f.write(cfgContent)
        # we replace the configuration by our own one.
        gConfig = ConfigurationClient(fileToLoadList=[self.testCfgFileName])
        self.setup = gConfig.getValue("/DIRAC/Setup", "")
        self.wm = gConfig.getValue("DIRAC/Setups/" + self.setup + "/WorkloadManagement", "")

    def tearDown(self):
        for aFile in [self.testCfgFileName, "pilot.json"]:
            try:
                os.remove(aFile)
            except OSError:
                pass
        self.clearCFG()

    @staticmethod
    def clearCFG():
        """SUPER UGLY: one must recreate the CFG objects of gConfigurationData
        not to conflict with other tests that might be using a local dirac.cfg"""
        gConfigurationData.localCFG = CFG()
        gConfigurationData.remoteCFG = CFG()
        gConfigurationData.mergedCFG = CFG()
        gConfigurationData.generateNewVersion()


class Test_PilotCStoJSONSynchronizer_sync(PilotCStoJSONSynchronizerTestCase):
    def test_success(self):
        synchroniser = PilotCStoJSONSynchronizer()
        res = synchroniser.getCSDict()
        assert res["OK"], res["Message"]
        res = synchroniser.getCSDict(includeMasterCS=False)
        assert res["OK"], res["Message"]


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(PilotCStoJSONSynchronizerTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(Test_PilotCStoJSONSynchronizer_sync))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" Test case for DIRAC.Core.Utilities.Network module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest

from DIRAC.Core.Utilities.Network import discoverInterfaces, getFQDN, getIPsForHostName, checkHostsMatch


def test_discoverInterfaces():
    interfaces = discoverInterfaces()
    assert len(interfaces) >= 1
    assert "lo" in interfaces
    for interfaceInfo in interfaces.values():
        assert "ip" in interfaceInfo
        assert "mac" in interfaceInfo


def test_getFQDN():
    assert isinstance(getFQDN(), str)


@pytest.mark.parametrize(
    "host1, host2, isValid, expected",
    [
        ("localhost", "localhost", True, True),
        ("localhost", "example.com", True, False),
        ("localhost", "example.invalid", False, False),
        ("example.com", "localhost", True, False),
        ("example.invalid", "localhost", False, False),
    ],
)
def test_checkHostsMatch(host1, host2, isValid, expected):
    result = checkHostsMatch(host1, host2)
    if isValid:
        assert result["OK"]
        assert result["Value"] is expected
    else:
        assert not result["OK"]


@pytest.mark.parametrize("hostname", ["localhost", "example.com"])
def test_getIPsForHostName(hostname):
    result = getIPsForHostName(hostname)
    assert result["OK"]
    assert len(result["Value"]) >= 1



#!/usr/bin/env python
########################################################################
# File :    dirac-install-web-portal
# Author :  Ricardo Graciani
########################################################################
"""
Do the initial installation of a DIRAC Web portal
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.disableCS()
    Script.parseCommandLine()

    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    gComponentInstaller.exitOnError = True
    gComponentInstaller.setupPortal()


if __name__ == "__main__":
    main()



"""
This module provides GOCDB2CSAgent code.

The agent is used to synchronize information between GOCDB and DIRAC configuration System (CS)
Right now it only adds perfosonar endpoints

The following options can be set for the GOCDB2CSAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN GOCDB2CSAgent
  :end-before: ##END
  :dedent: 2
  :caption: GOCDB2CSAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.LCG.GOCDBClient import GOCDBClient
from DIRAC.ConfigurationSystem.Client.Utilities import getDIRACGOCDictionary
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Config import gConfig


class GOCDB2CSAgent(AgentModule):
    """Class to retrieve information about service endpoints
    from GOCDB and update configuration stored by CS
    """

    def __init__(self, *args, **kwargs):
        """c'tor"""
        super(GOCDB2CSAgent, self).__init__(*args, **kwargs)
        self.GOCDBClient = None
        self.csAPI = None
        self.dryRun = False

    def initialize(self):
        """Run at the agent initialization (normally every 500 cycles)"""
        # client to connect to GOCDB
        self.GOCDBClient = GOCDBClient()
        self.dryRun = self.am_getOption("DryRun", self.dryRun)

        # API needed to update configuration stored by CS
        self.csAPI = CSAPI()
        return self.csAPI.initialize()

    def execute(self):
        """
        Execute GOCDB queries according to the function map
        and user request (options in configuration).
        """

        # __functionMap is at the end of the class definition
        for option, functionCall in GOCDB2CSAgent.__functionMap.items():
            optionValue = self.am_getOption(option, True)
            if optionValue:
                result = functionCall(self)
                if not result["OK"]:
                    self.log.error("%s() failed with message: %s" % (functionCall.__name__, result["Message"]))
                else:
                    self.log.info("Successfully executed %s" % functionCall.__name__)

        return S_OK()

    def updatePerfSONARConfiguration(self):
        """
        Get current status of perfSONAR endpoints from GOCDB
        and update CS configuration accordingly.
        """
        log = self.log.getSubLogger("updatePerfSONAREndpoints")
        log.debug("Begin function ...")

        # get endpoints
        result = self.__getPerfSONAREndpoints()
        if not result["OK"]:
            log.error("__getPerfSONAREndpoints() failed with message: %s" % result["Message"])
            return S_ERROR("Unable to fetch perfSONAR endpoints from GOCDB.")
        endpointList = result["Value"]

        # add DIRAC site name
        result = self.__addDIRACSiteName(endpointList)
        if not result["OK"]:
            log.error("__addDIRACSiteName() failed with message: %s" % result["Message"])
            return S_ERROR("Unable to extend the list with DIRAC site names.")
        extendedEndpointList = result["Value"]

        # prepare dictionary with new configuration
        result = self.__preparePerfSONARConfiguration(extendedEndpointList)
        if not result["OK"]:
            log.error("__preparePerfSONARConfiguration() failed with message: %s" % result["Message"])
            return S_ERROR("Unable to prepare a new perfSONAR configuration.")
        finalConfiguration = result["Value"]

        # update configuration according to the final status of endpoints
        self.__updateConfiguration(finalConfiguration)
        log.debug("Configuration updated succesfully")

        log.debug("End function.")
        return S_OK()

    def __getPerfSONAREndpoints(self):
        """
        Retrieve perfSONAR endpoint information directly from GOCDB.

        :return: List of perfSONAR endpoints (dictionaries) as stored by GOCDB.
        """

        log = self.log.getSubLogger("__getPerfSONAREndpoints")
        log.debug("Begin function ...")

        # get perfSONAR endpoints (latency and bandwidth) form GOCDB
        endpointList = []
        for endpointType in ["Latency", "Bandwidth"]:
            result = self.GOCDBClient.getServiceEndpointInfo("service_type", "net.perfSONAR.%s" % endpointType)

            if not result["OK"]:
                log.error("getServiceEndpointInfo() failed with message: %s" % result["Message"])
                return S_ERROR("Could not fetch %s endpoints from GOCDB" % endpointType.lower())

            log.debug("Number of %s endpoints: %s" % (endpointType.lower(), len(result["Value"])))
            endpointList.extend(result["Value"])

        log.debug("Number of perfSONAR endpoints: %s" % len(endpointList))
        log.debug("End function.")
        return S_OK(endpointList)

    def __preparePerfSONARConfiguration(self, endpointList):
        """
        Prepare a dictionary with a new CS configuration of perfSONAR endpoints.

        :return: Dictionary where keys are configuration paths (options and sections)
                 and values are values of corresponding options
                 or None in case of a path pointing to a section.
        """

        log = self.log.getSubLogger("__preparePerfSONARConfiguration")
        log.debug("Begin function ...")

        # static elements of a path
        rootPath = "/Resources/Sites"
        extPath = "Network"
        baseOptionName = "Enabled"
        options = {baseOptionName: "True", "ServiceType": "perfSONAR"}

        # enable GOCDB endpoints in configuration
        newConfiguration = {}
        for endpoint in endpointList:
            if endpoint["DIRACSITENAME"] is None:
                continue

            split = endpoint["DIRACSITENAME"].split(".")
            path = cfgPath(rootPath, split[0], endpoint["DIRACSITENAME"], extPath, endpoint["HOSTNAME"])
            for name, defaultValue in options.items():
                newConfiguration[cfgPath(path, name)] = defaultValue

        # get current configuration
        currentConfiguration = {}
        for option in options:
            result = gConfig.getConfigurationTree(rootPath, extPath + "/", "/" + option)
            if not result["OK"]:
                log.error("getConfigurationTree() failed with message: %s" % result["Message"])
                return S_ERROR("Unable to fetch perfSONAR endpoints from CS.")
            currentConfiguration.update(result["Value"])

        # disable endpoints that disappeared in GOCDB
        removedElements = set(currentConfiguration) - set(newConfiguration)
        newElements = set(newConfiguration) - set(currentConfiguration)

        addedEndpoints = int(len(newElements) / len(options))
        disabledEndpoints = 0
        for path in removedElements:
            if baseOptionName in path:
                newConfiguration[path] = "False"
                if currentConfiguration[path] != "False":
                    disabledEndpoints = disabledEndpoints + 1

        # inform what will be changed
        if addedEndpoints > 0:
            self.log.info("%s new perfSONAR endpoints will be added to the configuration" % addedEndpoints)

        if disabledEndpoints > 0:
            self.log.info("%s old perfSONAR endpoints will be disable in the configuration" % disabledEndpoints)

        if addedEndpoints == 0 and disabledEndpoints == 0:
            self.log.info("perfSONAR configuration is up-to-date")

        log.debug("End function.")
        return S_OK(newConfiguration)

    def __addDIRACSiteName(self, inputList):
        """
        Extend given list of GOCDB endpoints with DIRAC site name, i.e.
        add an entry "DIRACSITENAME" in dictionaries that describe endpoints.
        If given site name could not be found "DIRACSITENAME" is set to 'None'.

        :return: List of perfSONAR endpoints (dictionaries).
        """

        log = self.log.getSubLogger("__addDIRACSiteName")
        log.debug("Begin function ...")

        # get site name dictionary
        result = getDIRACGOCDictionary()
        if not result["OK"]:
            log.error("getDIRACGOCDictionary() failed with message: %s" % result["Message"])
            return S_ERROR("Could not get site name dictionary")

        # reverse the dictionary (assume 1 to 1 relation)
        DIRACGOCDict = result["Value"]
        GOCDIRACDict = dict(zip(DIRACGOCDict.values(), DIRACGOCDict))

        # add DIRAC site names
        outputList = []
        for entry in inputList:
            try:
                entry["DIRACSITENAME"] = GOCDIRACDict[entry["SITENAME"]]
            except KeyError:
                self.log.warn("No dictionary entry for %s. " % entry["SITENAME"])
                entry["DIRACSITENAME"] = None
            outputList.append(entry)

        log.debug("End function.")
        return S_OK(outputList)

    def __updateConfiguration(self, setElements=None, delElements=None):
        """
        Update configuration stored by CS.
        """
        if setElements is None:
            setElements = {}
        if delElements is None:
            delElements = []

        log = self.log.getSubLogger("__updateConfiguration")
        log.debug("Begin function ...")

        # assure existence and proper value of a section or an option
        for path, value in setElements.items():

            if value is None:
                section = path
            else:
                split = path.rsplit("/", 1)
                section = split[0]

            try:
                result = self.csAPI.createSection(section)
                if not result["OK"]:
                    log.error("createSection() failed with message: %s" % result["Message"])
            except Exception as e:
                log.error("Exception in createSection(): %s" % repr(e).replace(",)", ")"))

            if value is not None:
                try:
                    result = self.csAPI.setOption(path, value)
                    if not result["OK"]:
                        log.error("setOption() failed with message: %s" % result["Message"])
                except Exception as e:
                    log.error("Exception in setOption(): %s" % repr(e).replace(",)", ")"))

        # delete elements in the configuration
        for path in delElements:
            result = self.csAPI.delOption(path)
            if not result["OK"]:
                log.warn("delOption() failed with message: %s" % result["Message"])

                result = self.csAPI.delSection(path)
                if not result["OK"]:
                    log.warn("delSection() failed with message: %s" % result["Message"])

        if self.dryRun:
            log.info("Dry Run: CS won't be updated")
            self.csAPI.showDiff()
        else:
            # update configuration stored by CS
            result = self.csAPI.commit()
            if not result["OK"]:
                log.error("commit() failed with message: %s" % result["Message"])
                return S_ERROR("Could not commit changes to CS.")
            else:
                log.info("Committed changes to CS")

        log.debug("End function.")
        return S_OK()

    # define mapping between an agent option in the configuration and a function call
    __functionMap = {
        "UpdatePerfSONARS": updatePerfSONARConfiguration,
    }



""" The POOL XML Slice class provides a simple plugin module to create
    an XML file for applications to translate LFNs to TURLs. The input
    dictionary has LFNs as keys with all associated metadata as key,
    value pairs.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog

__RCSID__ = "$Id$"

COMPONENT_NAME = "PoolXMLSlice"


class PoolXMLSlice(object):

    #############################################################################
    def __init__(self, catalogName):
        """Standard constructor"""
        self.fileName = catalogName
        self.name = COMPONENT_NAME
        self.log = gLogger.getSubLogger(self.name)

    #############################################################################
    def execute(self, dataDict):
        """Given a dictionary of resolved input data, this will creates a POOL XML slice."""
        poolXMLCatName = self.fileName
        try:
            poolXMLCat = PoolXMLCatalog()
            self.log.verbose("Creating POOL XML slice")

            for lfn, mdataList in dataDict.items():
                # lfn,pfn,se,guid tuple taken by POOL XML Catalogue
                if not isinstance(mdataList, list):
                    mdataList = [mdataList]
                # As a file may have several replicas, set first the file, then the replicas
                poolXMLCat.addFile((lfn, None, None, mdataList[0]["guid"], None))
                for mdata in mdataList:
                    path = ""
                    if "path" in mdata:
                        path = mdata["path"]
                    elif os.path.exists(os.path.basename(mdata["pfn"])):
                        path = os.path.abspath(os.path.basename(mdata["pfn"]))
                    else:
                        path = mdata["turl"]
                    poolXMLCat.addReplica((lfn, path, mdata["se"], False))

            xmlSlice = poolXMLCat.toXML()
            self.log.verbose("POOL XML Slice is: ")
            self.log.verbose(xmlSlice)
            with open(poolXMLCatName, "w") as poolSlice:
                poolSlice.write(xmlSlice)
            self.log.info("POOL XML Catalogue slice written to %s" % (poolXMLCatName))
            try:
                # Temporary solution to the problem of storing the SE in the Pool XML slice
                with open("%s.temp" % (poolXMLCatName), "w") as poolSlice_temp:
                    xmlSlice = poolXMLCat.toXML(True)
                    poolSlice_temp.write(xmlSlice)
            except Exception as x:
                self.log.warn("Attempted to write catalog also to %s.temp but this failed" % (poolXMLCatName))
        except Exception as x:
            self.log.error(str(x))
            return S_ERROR("Exception during construction of POOL XML slice")

        return S_OK("POOL XML Slice created")


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" Threaded implementation of service interface
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from DIRAC import gLogger, S_OK

from DIRAC.ConfigurationSystem.private.ServiceInterfaceBase import ServiceInterfaceBase
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

__RCSID__ = "$Id$"


class ServiceInterface(ServiceInterfaceBase, threading.Thread):
    """
    Service interface, manage Slave/Master server for CS
    Thread components
    """

    def __init__(self, sURL):
        threading.Thread.__init__(self)
        ServiceInterfaceBase.__init__(self, sURL)

    def _launchCheckSlaves(self):
        """
        Start loop which check if slaves are alive
        """
        gLogger.info("Starting purge slaves thread")
        self.setDaemon(1)
        self.start()

    def run(self):
        while True:
            iWaitTime = gConfigurationData.getSlavesGraceTime()
            time.sleep(iWaitTime)
            self._checkSlavesStatus()

    def _updateServiceConfiguration(self, urlSet, fromMaster=False):
        """
        Update configuration of a set of slave services in parallel

        :param set urlSet: a set of service URLs
        :param fromMaster: flag to force updating from the master CS
        :return: Nothing
        """
        if not urlSet:
            return
        with ThreadPoolExecutor(max_workers=len(urlSet)) as executor:
            futureUpdate = {executor.submit(self._forceServiceUpdate, url, fromMaster): url for url in urlSet}
            for future in as_completed(futureUpdate):
                url = futureUpdate[future]
                result = future.result()
                if result["OK"]:
                    gLogger.info("Successfully updated slave configuration", url)
                else:
                    gLogger.error("Failed to update slave configuration", url)



""" UserProfileDB class is a front-end to the User Profile Database
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six

import cachetools

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Base.DB import DB


class UserProfileDB(DB):
    """UserProfileDB class is a front-end to the User Profile Database"""

    tableDict = {
        "up_Users": {
            "Fields": {
                "Id": "INTEGER AUTO_INCREMENT NOT NULL",
                "UserName": "VARCHAR(32) NOT NULL",
                "LastAccess": "DATETIME",
            },
            "PrimaryKey": "Id",
            "UniqueIndexes": {"U": ["UserName"]},
            "Engine": "InnoDB",
        },
        "up_Groups": {
            "Fields": {
                "Id": "INTEGER AUTO_INCREMENT NOT NULL",
                "UserGroup": "VARCHAR(32) NOT NULL",
                "LastAccess": "DATETIME",
            },
            "PrimaryKey": "Id",
            "UniqueIndexes": {"G": ["UserGroup"]},
            "Engine": "InnoDB",
        },
        "up_VOs": {
            "Fields": {
                "Id": "INTEGER AUTO_INCREMENT NOT NULL",
                "VO": "VARCHAR(32) NOT NULL",
                "LastAccess": "DATETIME",
            },
            "PrimaryKey": "Id",
            "UniqueIndexes": {"VO": ["VO"]},
            "Engine": "InnoDB",
        },
        "up_ProfilesData": {
            "Fields": {
                "UserId": "INTEGER",
                "GroupId": "INTEGER",
                "VOId": "INTEGER",
                "Profile": "VARCHAR(255) NOT NULL",
                "VarName": "VARCHAR(255) NOT NULL",
                "Data": "BLOB",
                "ReadAccess": 'VARCHAR(10) DEFAULT "USER"',
                "PublishAccess": 'VARCHAR(10) DEFAULT "USER"',
            },
            "PrimaryKey": ["UserId", "GroupId", "Profile", "VarName"],
            "Indexes": {
                "ProfileKey": ["UserId", "GroupId", "Profile"],
                "UserKey": ["UserId"],
            },
            "Engine": "InnoDB",
        },
    }

    def __init__(self):
        """Constructor"""
        self.__permValues = ["USER", "GROUP", "VO", "ALL"]
        self.__permAttrs = ["ReadAccess", "PublishAccess"]
        self.__cache = cachetools.TTLCache(1024, 15)
        DB.__init__(self, "UserProfileDB", "Framework/UserProfileDB")
        retVal = self.__initializeDB()
        if not retVal["OK"]:
            raise Exception("Can't create tables: %s" % retVal["Message"])

    def _checkTable(self):
        """Make sure the tables are created"""
        return self.__initializeDB()

    def __initializeDB(self):
        """
        Create the tables
        """
        retVal = self._query("show tables")
        if not retVal["OK"]:
            return retVal

        tablesInDB = [t[0] for t in retVal["Value"]]
        tablesD = {}

        if "up_Users" not in tablesInDB:
            tablesD["up_Users"] = self.tableDict["up_Users"]

        if "up_Groups" not in tablesInDB:
            tablesD["up_Groups"] = self.tableDict["up_Groups"]

        if "up_VOs" not in tablesInDB:
            tablesD["up_VOs"] = self.tableDict["up_VOs"]

        if "up_ProfilesData" not in tablesInDB:
            tablesD["up_ProfilesData"] = self.tableDict["up_ProfilesData"]

        return self._createTables(tablesD)

    def __getUserId(self, userName, insertIfMissing=True):
        return self.__getObjId(userName, "UserName", "up_Users", insertIfMissing)

    def __getGroupId(self, groupName, insertIfMissing=True):
        return self.__getObjId(groupName, "UserGroup", "up_Groups", insertIfMissing)

    def __getVOId(self, voName, insertIfMissing=True):
        return self.__getObjId(voName, "VO", "up_VOs", insertIfMissing)

    def __getFieldsCached(self, tableName, outFields, condDict):
        """Call getFields with a TTL cache

        The UserProfileDB is written in such a way that repeatedly makes the same
        DB queries thousands of times. To workaround this, use a simple short-lived
        TTL cache to dramatically improve performance.
        """
        key = (tableName, tuple(outFields), tuple(sorted(condDict.items())))
        if key in self.__cache:
            return self.__cache[key]
        result = self.getFields(tableName, outFields, condDict)
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data) > 0:
            objId = data[0][0]
            self.updateFields(tableName, ["LastAccess"], ["UTC_TIMESTAMP()"], {"Id": objId})
            self.__cache[key] = result
        return result

    def __getObjId(self, objValue, varName, tableName, insertIfMissing=True):
        result = self.__getFieldsCached(tableName, ["Id"], {varName: objValue})
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data) > 0:
            objId = data[0][0]
            return S_OK(objId)
        if not insertIfMissing:
            return S_ERROR("No entry %s for %s defined in the DB" % (objValue, varName))
        result = self.insertFields(tableName, [varName, "LastAccess"], [objValue, "UTC_TIMESTAMP()"])
        if not result["OK"]:
            return result
        return S_OK(result["lastRowId"])

    def getUserGroupIds(self, userName, userGroup, insertIfMissing=True):
        result = self.__getUserId(userName, insertIfMissing)
        if not result["OK"]:
            return result
        userId = result["Value"]
        result = self.__getGroupId(userGroup, insertIfMissing)
        if not result["OK"]:
            return result
        groupId = result["Value"]
        userVO = Registry.getVOForGroup(userGroup)
        if not userVO:
            userVO = "undefined"
        result = self.__getVOId(userVO, insertIfMissing)
        if not result["OK"]:
            return result
        voId = result["Value"]
        return S_OK((userId, groupId, voId))

    def deleteUserProfile(self, userName, userGroup=False):
        """
        Delete the profiles for a user
        """
        result = self.__getUserId(userName)
        if not result["OK"]:
            return result
        userId = result["Value"]
        condDict = {"UserId": userId}
        if userGroup:
            result = self.__getGroupId(userGroup)
            if not result["OK"]:
                return result
            groupId = result["Value"]
            condDict["GroupId"] = groupId
        result = self.deleteEntries("up_ProfilesData", condDict)
        if not result["OK"] or not userGroup:
            return result
        return self.deleteEntries("up_Users", {"Id": userId})

    def __webProfileUserDataCond(self, userIds, sqlProfileName=False, sqlVarName=False):
        condSQL = [
            "`up_ProfilesData`.UserId=%s" % userIds[0],
            "`up_ProfilesData`.GroupId=%s" % userIds[1],
            "`up_ProfilesData`.VOId=%s" % userIds[2],
        ]
        if sqlProfileName:
            condSQL.append("`up_ProfilesData`.Profile=%s" % sqlProfileName)
        if sqlVarName:
            condSQL.append("`up_ProfilesData`.VarName=%s" % sqlVarName)
        return " AND ".join(condSQL)

    def __webProfileReadAccessDataCond(self, userIds, ownerIds, sqlProfileName, sqlVarName=False, match=False):
        permCondSQL = []
        sqlCond = []

        if match:
            sqlCond.append(
                "`up_ProfilesData`.UserId = %s AND `up_ProfilesData`.GroupId = %s" % (ownerIds[0], ownerIds[1])
            )
        else:
            permCondSQL.append(
                "`up_ProfilesData`.UserId = %s AND `up_ProfilesData`.GroupId = %s" % (ownerIds[0], ownerIds[1])
            )

        permCondSQL.append('`up_ProfilesData`.GroupId=%s AND `up_ProfilesData`.ReadAccess="GROUP"' % userIds[1])
        permCondSQL.append('`up_ProfilesData`.VOId=%s AND `up_ProfilesData`.ReadAccess="VO"' % userIds[2])
        permCondSQL.append('`up_ProfilesData`.ReadAccess="ALL"')

        sqlCond.append("`up_ProfilesData`.Profile = %s" % sqlProfileName)
        if sqlVarName:
            sqlCond.append("`up_ProfilesData`.VarName = %s" % (sqlVarName))
        # Perms
        sqlCond.append("( ( %s ) )" % " ) OR ( ".join(permCondSQL))
        return " AND ".join(sqlCond)

    def __parsePerms(self, perms, addMissing=True):
        normPerms = {}
        for pName in self.__permAttrs:
            if not perms or pName not in perms:
                if addMissing:
                    normPerms[pName] = self.__permValues[0]
                continue
            permVal = perms[pName].upper()
            for nV in self.__permValues:
                if nV == permVal:
                    normPerms[pName] = nV
                    break
            if pName not in normPerms and addMissing:
                normPerms[pName] = self.__permValues[0]

        return normPerms

    def retrieveVarById(self, userIds, ownerIds, profileName, varName):
        """
        Get a data entry for a profile
        """
        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]

        result = self._escapeString(varName)
        if not result["OK"]:
            return result
        sqlVarName = result["Value"]

        sqlCond = self.__webProfileReadAccessDataCond(userIds, ownerIds, sqlProfileName, sqlVarName, True)
        # when we retrieve the user profile we have to take into account the user.
        selectSQL = "SELECT data FROM `up_ProfilesData` WHERE %s" % sqlCond
        result = self._query(selectSQL)
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data) > 0:
            return S_OK(data[0][0])
        return S_ERROR("No data for userIds %s profileName %s varName %s" % (userIds, profileName, varName))

    def retrieveAllUserVarsById(self, userIds, profileName):
        """
        Get a data entry for a profile
        """
        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]

        sqlCond = self.__webProfileUserDataCond(userIds, sqlProfileName)
        selectSQL = "SELECT varName, data FROM `up_ProfilesData` WHERE %s" % sqlCond
        result = self._query(selectSQL)
        if not result["OK"]:
            return result
        data = result["Value"]
        return S_OK(dict(data))

    def retrieveUserProfilesById(self, userIds):
        """
        Get all profiles and data for a user
        """
        sqlCond = self.__webProfileUserDataCond(userIds)
        selectSQL = "SELECT Profile, varName, data FROM `up_ProfilesData` WHERE %s" % sqlCond
        result = self._query(selectSQL)
        if not result["OK"]:
            return result
        data = result["Value"]
        dataDict = {}
        for row in data:
            if row[0] not in dataDict:
                dataDict[row[0]] = {}
            dataDict[row[0]][row[1]] = row[2]
        return S_OK(dataDict)

    def retrieveVarPermsById(self, userIds, ownerIds, profileName, varName):
        """
        Get a data entry for a profile
        """
        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]

        result = self._escapeString(varName)
        if not result["OK"]:
            return result
        sqlVarName = result["Value"]

        sqlCond = self.__webProfileReadAccessDataCond(userIds, ownerIds, sqlProfileName, sqlVarName)
        selectSQL = "SELECT %s FROM `up_ProfilesData` WHERE %s" % (", ".join(self.__permAttrs), sqlCond)
        result = self._query(selectSQL)
        if not result["OK"]:
            return result
        data = result["Value"]
        if len(data) > 0:
            permDict = {self.__permAttrs[i]: data[0][i] for i in range(len(self.__permAttrs))}
            return S_OK(permDict)
        return S_ERROR("No data for userIds %s profileName %s varName %s" % (userIds, profileName, varName))

    def deleteVarByUserId(self, userIds, profileName, varName):
        """
        Remove a data entry for a profile
        """
        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]

        result = self._escapeString(varName)
        if not result["OK"]:
            return result
        sqlVarName = result["Value"]

        sqlCond = self.__webProfileUserDataCond(userIds, sqlProfileName, sqlVarName)
        selectSQL = "DELETE FROM `up_ProfilesData` WHERE %s" % sqlCond
        return self._update(selectSQL)

    def storeVarByUserId(self, userIds, profileName, varName, data, perms):
        """
        Set a data entry for a profile
        """
        sqlInsertValues = []
        sqlInsertKeys = []

        sqlInsertKeys.append(("UserId", userIds[0]))
        sqlInsertKeys.append(("GroupId", userIds[1]))
        sqlInsertKeys.append(("VOId", userIds[2]))

        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]
        sqlInsertKeys.append(("Profile", sqlProfileName))

        result = self._escapeString(varName)
        if not result["OK"]:
            return result
        sqlVarName = result["Value"]
        sqlInsertKeys.append(("VarName", sqlVarName))

        result = self._escapeString(data)
        if not result["OK"]:
            return result
        sqlInsertValues.append(("Data", result["Value"]))

        normPerms = self.__parsePerms(perms)
        for k in normPerms:
            sqlInsertValues.append((k, '"%s"' % normPerms[k]))

        sqlInsert = sqlInsertKeys + sqlInsertValues
        insertSQL = "INSERT INTO `up_ProfilesData` ( %s ) VALUES ( %s )" % (
            ", ".join([f[0] for f in sqlInsert]),
            ", ".join([str(f[1]) for f in sqlInsert]),
        )
        result = self._update(insertSQL)
        if result["OK"]:
            return result
        # If error and not duplicate -> real error
        if "Duplicate entry" not in result["Message"]:
            return result
        updateSQL = "UPDATE `up_ProfilesData` SET %s WHERE %s" % (
            ", ".join(["%s=%s" % f for f in sqlInsertValues]),
            self.__webProfileUserDataCond(userIds, sqlProfileName, sqlVarName),
        )
        return self._update(updateSQL)

    def setUserVarPermsById(self, userIds, profileName, varName, perms):

        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]

        result = self._escapeString(varName)
        if not result["OK"]:
            return result
        sqlVarName = result["Value"]

        nPerms = self.__parsePerms(perms, False)
        if not nPerms:
            return S_OK()
        sqlPerms = ",".join("%s='%s'" % (k, nPerms[k]) for k in nPerms)

        updateSql = "UPDATE `up_ProfilesData` SET %s WHERE %s" % (
            sqlPerms,
            self.__webProfileUserDataCond(userIds, sqlProfileName, sqlVarName),
        )
        return self._update(updateSql)

    def retrieveVar(self, userName, userGroup, ownerName, ownerGroup, profileName, varName):
        """
        Get a data entry for a profile
        """
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]

        result = self.getUserGroupIds(ownerName, ownerGroup)
        if not result["OK"]:
            return result
        ownerIds = result["Value"]

        return self.retrieveVarById(userIds, ownerIds, profileName, varName)

    def retrieveUserProfiles(self, userName, userGroup):
        """
        Helper for getting data
        """
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.retrieveUserProfilesById(userIds)

    def retrieveAllUserVars(self, userName, userGroup, profileName):
        """
        Helper for getting data
        """
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.retrieveAllUserVarsById(userIds, profileName)

    def retrieveVarPerms(self, userName, userGroup, ownerName, ownerGroup, profileName, varName):
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]

        result = self.getUserGroupIds(ownerName, ownerGroup, False)
        if not result["OK"]:
            return result
        ownerIds = result["Value"]

        return self.retrieveVarPermsById(userIds, ownerIds, profileName, varName)

    def setUserVarPerms(self, userName, userGroup, profileName, varName, perms):
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.setUserVarPermsById(userIds, profileName, varName, perms)

    def storeVar(self, userName, userGroup, profileName, varName, data, perms=None):
        """
        Helper for setting data
        """
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.storeVarByUserId(userIds, profileName, varName, data, perms=perms)

    def deleteVar(self, userName, userGroup, profileName, varName):
        """
        Helper for deleting data
        """
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.deleteVarByUserId(userIds, profileName, varName)

    def __profilesCondGenerator(self, value, varType, initialValue=False):
        if isinstance(value, six.string_types):
            value = [value]
        ids = []
        if initialValue:
            ids.append(initialValue)
        for val in value:
            if varType == "user":
                result = self.__getUserId(val, insertIfMissing=False)
            elif varType == "group":
                result = self.__getGroupId(val, insertIfMissing=False)
            else:
                result = self.__getVOId(val, insertIfMissing=False)
            if not result["OK"]:
                continue
            ids.append(result["Value"])
        if varType == "user":
            fieldName = "UserId"
        elif varType == "group":
            fieldName = "GroupId"
        else:
            fieldName = "VOId"
        return "`up_ProfilesData`.%s in ( %s )" % (fieldName, ", ".join(str(iD) for iD in ids))

    def listVarsById(self, userIds, profileName, filterDict=None):
        result = self._escapeString(profileName)
        if not result["OK"]:
            return result
        sqlProfileName = result["Value"]
        sqlCond = [
            "`up_Users`.Id = `up_ProfilesData`.UserId",
            "`up_Groups`.Id = `up_ProfilesData`.GroupId",
            "`up_VOs`.Id = `up_ProfilesData`.VOId",
            self.__webProfileReadAccessDataCond(userIds, userIds, sqlProfileName),
        ]
        if filterDict:
            filterDict = {k.lower(): filterDict[k] for k in filterDict}
            for k in ("user", "group", "vo"):
                if k in filterDict:
                    sqlCond.append(self.__profilesCondGenerator(filterDict[k], k))

        sqlVars2Get = ["`up_Users`.UserName", "`up_Groups`.UserGroup", "`up_VOs`.VO", "`up_ProfilesData`.VarName"]
        sqlQuery = "SELECT %s FROM `up_Users`, `up_Groups`, `up_VOs`, `up_ProfilesData` WHERE %s" % (
            ", ".join(sqlVars2Get),
            " AND ".join(sqlCond),
        )

        return self._query(sqlQuery)

    def listVars(self, userName, userGroup, profileName, filterDict=None):
        result = self.getUserGroupIds(userName, userGroup)
        if not result["OK"]:
            return result
        userIds = result["Value"]
        return self.listVarsById(userIds, profileName, filterDict)

    def getUserProfileNames(self, permission):
        """
        it returns the available profile names by not taking account the permission: ReadAccess and PublishAccess
        """
        permissions = self.__parsePerms(permission, False)
        if not permissions:
            return S_OK([])

        condition = ",".join("%s='%s'" % (k, permissions[k]) for k in permissions)

        query = "SELECT distinct Profile from `up_ProfilesData` where %s" % condition
        retVal = self._query(query)
        if not retVal["OK"]:
            return retVal
        return S_OK([i[0] for i in retVal["Value"]])
